"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.context.context import Transforms
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.jsonschema._keywords_cfn import FnItems
from cfnlint.rules.functions.Ref import Ref
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = Ref()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {
            "Resources": {
                "MyVolume": {"Type": "AWS::EC2::Volume"},
                "MyInstance": {"Type": "AWS::EC2::Instance"},
            },
            "Parameters": {
                "MyParameter": {
                    "Type": "String",
                },
                "MyArrayParameter": {
                    "Type": "CommaDelimitedList",
                },
            },
        },
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,instance,schema,context_evolve,expected",
    [
        (
            "Valid ref",
            {"Ref": "MyVolume"},
            {"type": "string"},
            {},
            [],
        ),
        (
            "Valid array ref",
            {"Ref": "MyArrayParameter"},
            {"type": "array"},
            {},
            [],
        ),
        (
            "Invalid Ref with bad type",
            {"Ref": ["foo"]},
            {"type": "string"},
            {},
            [
                ValidationError(
                    "['foo'] is not of type 'string'",
                    path=deque(["Ref"]),
                    validator="ref",
                    schema_path=deque(["type"]),
                ),
                ValidationError(
                    (
                        "['foo'] is not one of ['MyParameter', 'MyArrayParameter', "
                        "'MyVolume', 'MyInstance', 'AWS::AccountId', "
                        "'AWS::NoValue', 'AWS::NotificationARNs', "
                        "'AWS::Partition', 'AWS::Region', "
                        "'AWS::StackId', 'AWS::StackName', 'AWS::URLSuffix']"
                    ),
                    path=deque(["Ref"]),
                    schema_path=deque(["enum"]),
                    validator="ref",
                ),
            ],
        ),
        (
            "Invalid Ref with a type that doesn't match array schema",
            {"Ref": "MyParameter"},
            {"type": "array"},
            {},
            [ValidationError("{'Ref': 'MyParameter'} is not of type 'array'")],
        ),
        (
            "Invalid Ref with a type that doesn't match singular schema",
            {"Ref": "Foo"},
            {"type": "string"},
            {},
            [
                ValidationError(
                    (
                        "'Foo' is not one of ['MyParameter', 'MyArrayParameter', "
                        "'MyVolume', 'MyInstance', 'AWS::AccountId', "
                        "'AWS::NoValue', 'AWS::NotificationARNs', "
                        "'AWS::Partition', 'AWS::Region', "
                        "'AWS::StackId', 'AWS::StackName', 'AWS::URLSuffix']"
                    ),
                    path=deque(["Ref"]),
                    schema_path=deque(["enum"]),
                    validator="ref",
                ),
            ],
        ),
        (
            "Valid ref",
            {"Ref": {"Ref": "MyParameter"}},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [],
        ),
        (
            "Invalid second ref with array return",
            {"Ref": {"Ref": "MyArrayParameter"}},
            {"type": "array"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [
                ValidationError(
                    "{'Ref': 'MyArrayParameter'} is not of type 'string'",
                    path=deque(["Ref"]),
                    validator="ref",
                    schema_path=deque(["ref"]),
                ),
            ],
        ),
    ],
)
def test_validate(name, instance, schema, context_evolve, expected, rule, context, cfn):
    context = context.evolve(**context_evolve)
    validator = CfnTemplateValidator({}).extend(
        validators={
            "fn_items": FnItems().validate,
            "ref": Ref().ref,
        }
    )(context=context, cfn=cfn)
    errs = list(rule.ref(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class SnapStart(CfnLintKeyword):
    """Check if Lambda SnapStart is properly configured"""

    id = "W2530"
    shortdesc = "Validate that SnapStart is properly configured"
    description = (
        "To properly leverage SnapStart, you must configure both the lambda function "
        "and attach a Lambda version resource"
    )
    source_url = "https://docs.aws.amazon.com/lambda/latest/dg/snapstart.html"
    tags = ["resources", "lambda"]

    def __init__(self):
        super().__init__(
            ["Resources/AWS::Lambda::Function/Properties/SnapStart/ApplyOn"]
        )

    def validate(
        self, validator: Validator, _, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not validator.is_type(instance, "string"):
            return

        if instance != "PublishedVersions":
            return

        resource_name: str = str(validator.context.path.path[1])
        lambda_version_type = "AWS::Lambda::Version"
        if list(
            validator.cfn.get_resource_children(resource_name, [lambda_version_type])
        ):
            return

        yield ValidationError(
            (
                f"'SnapStart' is enabled but an {lambda_version_type!r} "
                "resource is not attached"
            ),
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any, Iterator

from cfnlint.helpers import ensure_list, is_function
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.helpers import get_resource_by_name, get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class InstanceImageId(CfnLintKeyword):
    id = "E3673"
    shortdesc = "Validate if an ImageId is required"
    description = (
        "Validate if an ImageID is required. It can be "
        "required if the associated LaunchTemplate doesn't specify "
        "an ImageID"
    )
    tags = ["resources", "ec2"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::EC2::Instance/Properties",
            ],
        )

    def _get_related_launch_template(
        self, validator: Validator, instance: Any
    ) -> Iterator[tuple[Any, Validator]]:

        for launch_template, launch_template_validator in get_value_from_path(
            validator, instance, deque(["LaunchTemplate"])
        ):
            if not launch_template:
                continue
            for id, id_validator in get_value_from_path(
                launch_template_validator, launch_template, deque(["LaunchTemplateId"])
            ):
                if id is None:
                    for name, name_validator in get_value_from_path(
                        id_validator,
                        launch_template,
                        deque(["LaunchTemplateName"]),
                    ):
                        yield name, name_validator
                yield id, id_validator

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        for (
            instance_image_id,
            instance_image_id_validator,
        ) in get_value_from_path(
            validator,
            instance,
            path=deque(["ImageId"]),
        ):
            if instance_image_id:
                continue

            launch_templates = list(
                self._get_related_launch_template(instance_image_id_validator, instance)
            )
            path: deque[str | int] = deque([])
            if "ImageId" != instance_image_id_validator.context.path.path[-1]:
                path = deque(
                    list(instance_image_id_validator.context.path.path)[
                        len(validator.context.path.path) :
                    ]
                )

            if not launch_templates:
                yield ValidationError(
                    "'ImageId' is a required property",
                    validator="required",
                    path=path,
                    rule=self,
                )
                continue

            for (
                instance_launch_template,
                instance_launch_template_validator,
            ) in launch_templates:
                fn_k, fn_v = is_function(instance_launch_template)

                # if its not a function we can't tell from a string
                # if the image ID is there or not
                if fn_k is None:
                    continue

                launch_template, launch_template_validator = None, None
                if fn_k == "Ref":
                    if fn_v in instance_launch_template_validator.context.parameters:
                        continue
                    elif fn_v in instance_launch_template_validator.context.resources:
                        launch_template, launch_template_validator = (
                            get_resource_by_name(
                                instance_launch_template_validator,
                                fn_v,
                                ["AWS::EC2::LaunchTemplate"],
                            )
                        )
                    else:
                        continue
                elif fn_k == "Fn::GetAtt":
                    launch_template, launch_template_validator = get_resource_by_name(
                        instance_launch_template_validator,
                        ensure_list(fn_v)[0].split(".")[0],
                        ["AWS::EC2::LaunchTemplate"],
                    )
                else:
                    continue

                for (
                    launch_template_image_id,
                    _,
                ) in get_value_from_path(
                    launch_template_validator,
                    launch_template or {},
                    path=deque(["Properties", "LaunchTemplateData", "ImageId"]),
                ):
                    if launch_template_image_id is None and instance_image_id is None:
                        yield ValidationError(
                            "'ImageId' is a required property",
                            validator="required",
                            path=path,
                            rule=self,
                        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.Select import Select
from cfnlint.rules.functions.SelectResolved import SelectResolved


@pytest.fixture(scope="module")
def rule():
    rule = Select()
    rule.child_rules["W1035"] = SelectResolved()
    yield rule


@pytest.fixture
def template():
    return {"Transform": "AWS::LanguageExtensions"}


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Invalid Fn::Select using an invalid function for index",
            {"Fn::Select": [{"Fn::GetAtt": "MyResource"}, ["bar"]]},
            {"type": "string"},
            [
                ValidationError(
                    "{'Fn::GetAtt': 'MyResource'} is not of type 'integer'",
                    path=deque(["Fn::Select", 0]),
                    schema_path=deque(["fn_items", "type"]),
                    validator="fn_select",
                ),
            ],
        ),
        (
            "Valid Fn::Length with language extension",
            {"Fn::Select": [{"Fn::Length": [1, 2]}, ["A", "B", "C"]]},
            {"type": "string"},
            [],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.fn_select(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_cloudwatch_alarm
from cfnlint.jsonschema.exceptions import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class AlarmPeriod(CfnLintJsonSchema):
    id = "E3615"
    shortdesc = "Validate the period is a valid value"
    description = "Valid values are 10, 30, 60, and any multiple of 60."
    tags = ["resources", "cloudwatch"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::CloudWatch::Alarm/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_cloudwatch_alarm,
                filename="period.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return f"{err.instance!r} is not one of [10, 30, 60] or a multiple of 60"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import datetime
from test.testlib.testcase import BaseTestCase

from cfnlint.helpers import format_json_string


class TestFormatJson(BaseTestCase):
    """Test Dumping JSON objects"""

    def test_success_run(self):
        """Test success run"""
        obj = {
            "Key": {"SubKey": "Value", "Date": datetime.datetime(2010, 9, 9)},
            "List": [{"SubKey": "AnotherValue"}],
        }
        success = """{
 "Key": {
  "Date": "2010-09-09 00:00:00",
  "SubKey": "Value"
 },
 "List": [
  {
   "SubKey": "AnotherValue"
  }
 ]
}"""
        results = format_json_string(obj)
        self.assertEqual(success, results)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque

import jsonpatch
import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.resources.ecs.ServiceDynamicPorts import ServiceDynamicPorts


@pytest.fixture
def rule():
    rule = ServiceDynamicPorts()
    yield rule


_task_definition = {
    "Type": "AWS::ECS::TaskDefinition",
    "Properties": {
        "ContainerDefinitions": [
            {
                "Name": "my-container",
                "Image": "my-image",
                "PortMappings": [
                    {
                        "ContainerPort": 8080,
                        "HostPort": 0,
                    }
                ],
            }
        ],
    },
}

_target_group = {
    "Type": "AWS::ElasticLoadBalancingV2::TargetGroup",
    "Properties": {
        "HealthCheckPort": "traffic-port",
        "Port": 8080,
        "Protocol": "HTTP",
    },
}

_service = {
    "Type": "AWS::ECS::Service",
    "Properties": {
        "TaskDefinition": {"Ref": "TaskDefinition"},
        "LoadBalancers": [
            {
                "TargetGroupArn": {"Ref": "TargetGroup"},
                "ContainerName": "my-container",
                "ContainerPort": 8080,
            }
        ],
    },
}


@pytest.mark.parametrize(
    "template,start_path,expected",
    [
        (
            {
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "TargetGroup": dict(_target_group),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/HealthCheckPort",
                                "value": "3000",
                            },
                        ],
                    ),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Parameters": {
                    "MyHealthCheckPort": {
                        "Type": "String",
                    },
                },
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/HealthCheckPort",
                                "value": {"Ref": "MyHealthCheckPort"},
                            },
                        ],
                    ),
                    "Service": dict(_service),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Parameters": {
                    "MyContainerPort": {
                        "Type": "String",
                    },
                },
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "replace",
                                "path": (
                                    "/Properties/ContainerDefinitions"
                                    "/0/PortMappings/0/ContainerPort"
                                ),
                                "value": {"Ref": "MyContainerPort"},
                            },
                        ],
                    ),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/HealthCheckPort",
                                "value": "3000",
                            },
                        ],
                    ),
                    "Service": dict(_service),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Parameters": {
                    "MyTaskDefinition": {
                        "Type": "String",
                    },
                },
                "Resources": {
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/HealthCheckPort",
                                "value": "3000",
                            },
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/TaskDefinition",
                                "value": {"Ref": "MyTaskDefinition"},
                            },
                        ],
                    ),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/HealthCheckPort",
                                "value": "3000",
                            },
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/TaskDefinition",
                                "value": {"Fn::FindInMap": ["A", "B", "C"]},
                            },
                        ],
                    ),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/HealthCheckPort",
                                "value": "3000",
                            },
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/TaskDefinition",
                                "value": {"Fn::GetAtt": ["TaskDefinition", "Arn"]},
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [
                ValidationError(
                    (
                        "When using an ECS task definition of host "
                        "port 0 and associating that container to "
                        "an ELB the target group has to have a "
                        "'HealthCheckPort' of 'traffic-port'"
                    ),
                    validator="const",
                    path_override=deque(
                        ["Resources", "TargetGroup", "Properties", "HealthCheckPort"]
                    ),
                    rule=ServiceDynamicPorts(),
                )
            ],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/HealthCheckPort",
                                "value": "3000",
                            },
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/TaskDefinition",
                                "value": {"Ref": []},
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/HealthCheckPort",
                                "value": "3000",
                            },
                        ],
                    ),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [
                ValidationError(
                    (
                        "When using an ECS task definition of host "
                        "port 0 and associating that container to "
                        "an ELB the target group has to have a "
                        "'HealthCheckPort' of 'traffic-port'"
                    ),
                    validator="const",
                    path_override=deque(
                        ["Resources", "TargetGroup", "Properties", "HealthCheckPort"]
                    ),
                    rule=ServiceDynamicPorts(),
                )
            ],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {"op": "remove", "path": "/Properties/HealthCheckPort"},
                        ],
                    ),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [
                ValidationError(
                    (
                        "When using an ECS task definition of host "
                        "port 0 and associating that container to "
                        "an ELB the target group has to have a "
                        "'HealthCheckPort' of 'traffic-port'"
                    ),
                    validator="required",
                    path_override=deque(
                        ["Resources", "TargetGroup", "Properties", "HealthCheckPort"]
                    ),
                    rule=ServiceDynamicPorts(),
                )
            ],
        ),
        (
            {
                "Parameters": {
                    "MyContainerName": {
                        "Type": "String",
                    }
                },
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/ContainerDefinitions/0/Name",
                                "value": {"Ref": "MyContainerName"},
                            },
                        ],
                    ),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {"op": "remove", "path": "/Properties/HealthCheckPort"},
                        ],
                    ),
                    "Service": dict(_service),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Parameters": {
                    "MyContainerName": {
                        "Type": "String",
                    }
                },
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {"op": "remove", "path": "/Properties/HealthCheckPort"},
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/LoadBalancers/0/ContainerName",
                                "value": {"Ref": "MyContainerName"},
                            },
                        ],
                    ),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Parameters": {
                    "MyContainerName": {
                        "Type": "String",
                    }
                },
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {"op": "remove", "path": "/Properties/HealthCheckPort"},
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "remove",
                                "path": "/Properties/LoadBalancers/0/ContainerName",
                            },
                        ],
                    ),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {"op": "remove", "path": "/Properties/HealthCheckPort"},
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "remove",
                                "path": "/Properties/LoadBalancers/0/ContainerPort",
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {"op": "remove", "path": "/Properties/HealthCheckPort"},
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "remove",
                                "path": "/Properties/LoadBalancers/0/TargetGroupArn",
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/ContainerDefinitions/0/Name",
                                "value": "Changed",
                            },
                        ],
                    ),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {"op": "remove", "path": "/Properties/HealthCheckPort"},
                        ],
                    ),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "replace",
                                "path": (
                                    "/Properties/ContainerDefinitions/"
                                    "0/PortMappings/0/ContainerPort"
                                ),
                                "value": "30",
                            },
                        ],
                    ),
                    "TargetGroup": jsonpatch.apply_patch(
                        dict(_target_group),
                        [
                            {"op": "remove", "path": "/Properties/HealthCheckPort"},
                        ],
                    ),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "replace",
                                "path": (
                                    "/Properties/ContainerDefinitions/"
                                    "0/PortMappings/0/HostPort"
                                ),
                                "value": "30",
                            },
                        ],
                    ),
                    "TargetGroup": dict(_target_group),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
    ],
    indirect=["template"],
)
def test_validate(template, start_path, expected, rule, validator):
    for instance, instance_validator in get_value_from_path(
        validator, template, start_path
    ):
        errs = list(rule.validate(instance_validator, "", instance, {}))
        assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.rds.DbInstanceAuroraExclusive import (
    DbInstanceAuroraExclusive,
)


@pytest.fixture(scope="module")
def rule():
    rule = DbInstanceAuroraExclusive()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "Engine": "aurora",
            },
            [],
        ),
        (
            {
                "Engine": "Auora",
            },
            [],
        ),
        (
            {"MasterUserPassword": "Test"},
            [],
        ),
        (
            {
                "Engine": "aurora",
                "DeletionProtection": True,
            },
            [
                ValidationError(
                    "Additional properties are not allowed (['DeletionProtection'])",
                    rule=DbInstanceAuroraExclusive(),
                    path=deque([]),
                    instance={"Engine": "aurora", "DeletionProtection": True},
                    validator="not",
                    schema_path=deque(["then", "not"]),
                )
            ],
        ),
        (
            {
                "Engine": "Aurora",
                "DeletionProtection": True,
            },
            [
                ValidationError(
                    "Additional properties are not allowed (['DeletionProtection'])",
                    rule=DbInstanceAuroraExclusive(),
                    path=deque([]),
                    instance={"Engine": "aurora", "DeletionProtection": True},
                    validator="not",
                    schema_path=deque(["then", "not"]),
                )
            ],
        ),
        (
            {
                "Engine": "aurora",
                "DeletionProtection": True,
                "CopyTagsToSnapshot": True,
            },
            [
                ValidationError(
                    (
                        "Additional properties are not "
                        "allowed (['DeletionProtection', "
                        "'CopyTagsToSnapshot'])"
                    ),
                    rule=DbInstanceAuroraExclusive(),
                    path=deque([]),
                    instance={
                        "Engine": "aurora",
                        "DeletionProtection": True,
                        "CopyTagsToSnapshot": True,
                    },
                    validator="not",
                    schema_path=deque(["then", "not"]),
                )
            ],
        ),
        (
            [],
            [],
        ),
    ],
)
def test_backup_lifecycle(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import cfnlint.data.schemas.extensions.aws_dynamodb_table
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class TableSseSpecification(CfnLintJsonSchema):
    id = "E3640"
    shortdesc = (
        "Validate DynamoDB SSE Specification has required properties when using KMS"
    )
    description = (
        "When doing KMS encryption in an AWS DynamoDB "
        "table there are required properties."
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::DynamoDB::Table/Properties/SSESpecification"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_dynamodb_table,
                filename="ssespecification_kms.json",
            ),
            all_matches=True,
        )



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from functools import cmp_to_key
from typing import Any

from sympy import Symbol

from cfnlint.conditions._utils import get_hash
from cfnlint.helpers import is_function

LOGGER = logging.getLogger(__name__)
REF_REGION = get_hash({"Ref": "AWS::Region"})


def _sorted(a: dict[str, str] | str, b: dict[str, str] | str) -> int:
    if isinstance(a, dict) and isinstance(b, dict):
        a_k, a_v = is_function(a)
        b_k, b_v = is_function(b)
        if a_k is None or b_k is None:
            if a_k is None and b_k is not None:
                return 1
            if a_k is not None and b_k is None:
                return -1
            return 0

        if a_k == "Ref" and isinstance(a_v, str):
            if b_k == "Ref" and isinstance(b_v, str):
                return -1 if a_v <= b_v else 1

        return -1 if a_k <= b_k else 1

    return -1 if str(a) <= str(b) else 1


_equals_cmp_key = cmp_to_key(_sorted)


@dataclass(frozen=True)
class EqualParameter:
    instance: dict[str, str] | str | None = field(init=True)
    hash: str = field(init=False)
    is_region: bool = field(init=False, default=False)

    def __post_init__(self):
        object.__setattr__(self, "hash", get_hash(self.instance))
        object.__setattr__(self, "is_region", self.hash == REF_REGION)

    @classmethod
    def create_from_instance(cls, instance: Any) -> "EqualParameter":
        if isinstance(instance, (str, bool, int, float)):
            return cls(instance=str(instance))

        if isinstance(instance, dict):
            return cls(instance=instance)

        # escape version
        if instance is None:
            return cls(instance=None)

        raise ValueError("EqualParameter has to be a string or a dict")

    def __eq__(self, __o: Any):
        if isinstance(__o, EqualParameter):
            return self.hash == __o.hash
        return False


@dataclass(frozen=True)
class Equal:
    instance: list[str | dict] = field(init=True)
    hash: str = field(init=False)

    left: EqualParameter = field(init=True)
    right: EqualParameter = field(init=True)

    is_static: bool | None = field(init=False, default=None)

    cnf: Symbol = field(init=False)

    @classmethod
    def create_from_instance(cls, instance: Any) -> "Equal":
        if not (isinstance(instance, list) and len(instance) == 2):
            raise ValueError("Equals has to be a list of two values")

        instance.sort(key=_equals_cmp_key)
        left = EqualParameter.create_from_instance(instance[0])
        right = EqualParameter.create_from_instance(instance[1])
        return cls(instance=instance, left=left, right=right)

    def __post_init__(self):
        object.__setattr__(self, "hash", get_hash(self.instance))
        if isinstance(self.left.instance, (str, int, bool, float)) and isinstance(
            self.right.instance, (str, int, bool, float)
        ):
            object.__setattr__(
                self, "is_static", self.left.instance == self.right.instance
            )
        object.__setattr__(self, "cnf", Symbol(self.hash))

    @property
    def parameters(self) -> list[EqualParameter]:
        """Returns a List of the EqualParameter that make up the Condition

        Args: None

        Returns:
            list[Equal]: A list of the left and right equal parameters if they are
                         of type EqualParameter
        """
        return [self.left, self.right]

    @property
    def is_region(self) -> bool:
        """
         Returns a bool if the condition is comparing a region to a string

        Args: None

        Returns:
            bool: If a parameter in the equals has {"Ref": "AWS::Region}
        """
        return (self.left.is_region or self.right.is_region) and not self.is_static



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import string
from unittest import TestCase

from cfnlint.conditions import UnknownSatisfisfaction
from cfnlint.conditions._utils import get_hash
from cfnlint.decode import decode_str
from cfnlint.template import Template


class TestConditions(TestCase):
    """Test Conditions"""

    def test_bad_condition_definition(self):
        """Badly formatted condition statements will return no results"""
        template = decode_str(
            """
        Conditions:
          IsProd: !Equals [!Ref Environment, "prod", "production"]
          IsUsEast1: !Equals [!Ref "AWS::Region", "us-east-1"]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(
            len(cfn.conditions._conditions), 1
        )  # would be 2 but IsProd fails
        # test coverage for KeyErrors in the following functions
        self.assertTrue(
            cfn.conditions.check_implies(
                {
                    "Test": True,
                },
                "IsUsEast1",
            )
        )
        self.assertEqual(
            list(cfn.conditions.build_scenarios({"IsProd": None, "IsUsEast1": None})),
            [],
        )

    def test_run_away_scenarios(self):
        """We cap runaway scenarios"""
        template = {
            "Parameters": {},
            "Conditions": {},
        }
        condition_names = {}
        for p in string.ascii_letters[0:10]:
            template["Parameters"][f"{p}Parameter"] = {
                "Type": "String",
            }
            template["Conditions"][f"{p}Condition"] = {
                "Fn::Equals": [{"Ref": f"{p}Parameter"}, "{p}"]
            }
            condition_names[f"{p}Condition"] = None

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 10)
        self.assertEqual(
            len(list(cfn.conditions.build_scenarios(condition_names))),
            cfn.conditions._max_scenarios,
        )

    def test_check_implies(self):
        """We properly validate implies scenarios"""
        template = {
            "Parameters": {},
            "Conditions": {},
        }
        condition_names = []
        for p in string.ascii_letters[0:4]:
            template["Parameters"][f"{p}Parameter"] = {
                "Type": "String",
            }
            template["Conditions"][f"{p}Condition"] = {
                "Fn::Equals": [{"Ref": f"{p}Parameter"}, "{p}"]
            }
            condition_names.append(f"{p}Condition")

        cfn = Template("", template)
        self.assertFalse(
            cfn.conditions.check_implies({"aCondition": False}, "aCondition")
        )
        self.assertTrue(
            cfn.conditions.check_implies({"aCondition": True}, "aCondition")
        )
        self.assertTrue(
            cfn.conditions.check_implies(
                {"aCondition": True, "bCondition": False}, "aCondition"
            )
        )

    def test_check_always_true_or_false(self):
        """We properly validate static equals"""
        template = decode_str(
            """
        Parameters:
          FalseParameter:
            Default: "false"
            Type: String
        Conditions:
          IsTrue: !Equals ["true", "true"]
          IsFalse: !Equals [!Ref FalseParameter, !Ref FalseParameter]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 2)
        # test coverage for KeyErrors in the following functions
        self.assertTrue(cfn.conditions.check_implies({"IsTrue": True}, "IsFalse"))

    def test_check_never_false(self):
        """With allowed values two conditions can not both be false"""
        template = decode_str(
            """
        Parameters:
          Environment:
            Type: String
            AllowedValues: ["prod", "dev"]
        Conditions:
          IsProd: !Equals [!Ref Environment, "prod"]
          IsDev: !Equals [!Ref Environment, "dev"]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 2)
        self.assertListEqual(
            list(cfn.conditions.build_scenarios({"IsProd": None, "IsDev": None})),
            [
                {"IsProd": True, "IsDev": False},
                {"IsProd": False, "IsDev": True},
            ],
        )

    def test_check_can_be_false(self):
        """With allowed values two conditions can both be false"""
        template = decode_str(
            """
        Parameters:
          Environment:
            Type: String
            AllowedValues: ["prod", "dev", "stage"]
        Conditions:
          IsProd: !Equals [!Ref Environment, "prod"]
          IsDev: !Equals [!Ref Environment, "dev"]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 2)
        self.assertListEqual(
            list(cfn.conditions.build_scenarios({"IsProd": None, "IsDev": None})),
            [
                {"IsProd": True, "IsDev": False},
                {"IsProd": False, "IsDev": True},
                {"IsProd": False, "IsDev": False},
            ],
        )
        self.assertListEqual(
            list(cfn.conditions.build_scenarios({"IsProd": {True}, "IsDev": None})),
            [
                {"IsProd": True, "IsDev": False},
            ],
        )
        self.assertListEqual(
            list(cfn.conditions.build_scenarios({"IsProd": {False}, "IsDev": None})),
            [
                {"IsProd": False, "IsDev": True},
                {"IsProd": False, "IsDev": False},
            ],
        )

    def test_check_can_be_good_when_condition_value(self):
        """Some times a condition Equals doesn't match to allowed values"""
        template = decode_str(
            """
        Parameters:
          Environment:
            Type: String
            AllowedValues: ["prod", "dev", "stage"]
        Conditions:
          IsGamma: !Equals [!Ref Environment, "gamma"]
          IsBeta: !Equals ["beta", !Ref Environment]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 2)
        self.assertListEqual(
            list(cfn.conditions.build_scenarios({"IsGamma": None, "IsBeta": None})),
            [
                {"IsBeta": False, "IsGamma": False},
            ],
        )
        self.assertListEqual(
            list(cfn.conditions.build_scenarios({"IsGamma": None})),
            [
                {"IsGamma": False},
            ],
        )

    def test_check_condition_region(self):
        """Regional based condition testing"""
        template = decode_str(
            """
        Parameters:
          Environment:
            Type: String
            AllowedValues: ["prod", "dev", "stage"]
        Conditions:
          IsUsEast1: !Equals [!Ref AWS::Region, "us-east-1"]
          IsUsWest2: !Equals ["us-west-2", !Ref AWS::Region]
          IsProd: !Equals [!Ref Environment, "prod"]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 3)
        self.assertListEqual(
            cfn.conditions.build_scenerios_on_region("IsUsEast1", "us-east-1"),
            [
                True,
            ],
        )
        self.assertListEqual(
            cfn.conditions.build_scenerios_on_region("IsUsEast1", "us-west-2"),
            [
                False,
            ],
        )
        self.assertListEqual(
            cfn.conditions.build_scenerios_on_region("IsUsWest2", "us-west-2"),
            [
                True,
            ],
        )
        self.assertListEqual(
            cfn.conditions.build_scenerios_on_region("IsUsWest2", "us-east-1"),
            [
                False,
            ],
        )
        self.assertListEqual(
            cfn.conditions.build_scenerios_on_region("IsProd", "us-east-1"),
            [
                True,
                False,
            ],
        )
        self.assertListEqual(
            cfn.conditions.build_scenerios_on_region("Foo", "us-east-1"),
            [
                True,
                False,
            ],
        )

    def test_test_condition(self):
        """Get condition and test"""
        template = decode_str(
            """
        Parameters:
          Environment:
            Type: String
            AllowedValues: ["prod", "dev", "stage"]
        Conditions:
          IsUsEast1: !Equals [!Ref AWS::Region, "us-east-1"]
          IsUsWest2: !Equals ["us-west-2", !Ref AWS::Region]
          IsProd: !Equals [!Ref Environment, "prod"]
          IsUsEast1AndProd: !And [!Condition IsUsEast1, !Condition IsProd]
        """
        )[0]

        h_region = get_hash({"Ref": "AWS::Region"})
        h_environment = get_hash({"Ref": "Environment"})

        cfn = Template("", template)
        self.assertTrue(cfn.conditions.get("IsUsEast1").test({h_region: "us-east-1"}))
        self.assertFalse(cfn.conditions.get("IsProd").test({h_environment: "dev"}))
        self.assertTrue(
            cfn.conditions.get("IsUsEast1AndProd").test(
                {h_region: "us-east-1", h_environment: "prod"}
            )
        )
        self.assertFalse(
            cfn.conditions.get("IsUsEast1AndProd").test(
                {h_region: "us-east-1", h_environment: "dev"}
            )
        )

    def test_build_scenerios_on_region_with_condition_dne(self):
        """Get condition and test"""
        template = decode_str(
            """
        Conditions:
          IsUsEast1: !Equals [!Ref AWS::Region, "us-east-1"]
        """
        )[0]

        cfn = Template("", template)
        self.assertListEqual(
            list(cfn.conditions.build_scenerios_on_region("IsProd", "us-east-1")),
            [True, False],
        )

    def test_satifaction(self):
        """Get condition and test"""
        template = decode_str(
            """
        Parameters:
          SecurityGroups:
            Default: ""
            Type: CommaDelimitedList
        Conditions:
          IsUsEast1: !Equals [!Ref AWS::Region, "us-east-1"]
          HasSecurityGroups: !Not [ !Equals [ !Join [ '', !Ref SecurityGroups ], ''] ]
        """
        )[0]

        cfn = Template("", template)

        with self.assertRaises(UnknownSatisfisfaction):
            cfn.conditions.satisfiable(
                {"HasSecurityGroups": True}, {"SecurityGroups": [""]}
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import os
from test.testlib.testcase import BaseTestCase

from cfnlint.config import _DEFAULT_RULESDIR
from cfnlint.helpers import load_plugins
from cfnlint.rules import CloudFormationLintRule


class TestLoadPlugins(BaseTestCase):
    """Test loading rules."""

    def testFromDefaultDirectory(self):
        rules = load_plugins(_DEFAULT_RULESDIR)

        self.assertTrue(all(isinstance(r, CloudFormationLintRule) for r in rules))
        # From templates/Base.py
        self.assertTrue("E1001" in (r.id for r in rules))
        # From resources/Name.py
        self.assertTrue("E3006" in (r.id for r in rules))

    def testFromSubDirectory(self):
        path = os.path.join(_DEFAULT_RULESDIR, "templates")
        rules = load_plugins(path)

        self.assertTrue(all(isinstance(r, CloudFormationLintRule) for r in rules))
        self.assertTrue("E1002" in (r.id for r in rules), rules)
        self.assertFalse("E3006" in (r.id for r in rules), rules)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.helpers import FUNCTIONS, TEMPLATED_PROPERTY_CFN_PATHS
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class PropertiesTemplated(CfnLintKeyword):
    """Check Base Resource Configuration"""

    id = "W3002"
    shortdesc = (
        "Warn when properties are configured to only work with the package command"
    )
    description = (
        "Some properties can be configured to only work with the CloudFormation"
        "package command. Warn when this is the case so user is aware."
    )
    source_url = (
        "https://docs.aws.amazon.com/cli/latest/reference/cloudformation/package.html"
    )
    tags = ["resources"]

    def __init__(self):
        """Init"""
        super().__init__(TEMPLATED_PROPERTY_CFN_PATHS)

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not isinstance(instance, str):
            return

        if validator.cfn.has_serverless_transform():
            return

        if validator.context.path.path[-1] in FUNCTIONS:
            return

        if not instance.startswith("s3://") and not instance.startswith("https://"):
            yield ValidationError("This code may only work with 'package' cli command")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from unittest.mock import MagicMock

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.context.context import Resource, Transforms
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.functions.FindInMap import FindInMap
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = FindInMap()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {
            "Parameters": {
                "MyParameter": {
                    "Type": "String",
                    "AllowedValues": ["A", "B", "C"],
                }
            },
            "Resources": {"MyResource": Resource({"Type": "AWS::SSM::Parameter"})},
            "Mappings": {"A": {"B": {"C": "Value"}}},
        },
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,instance,schema,context_evolve,ref_mock_values,expected",
    [
        (
            "Valid Fn::FindInMap",
            {"Fn::FindInMap": ["A", "B", "C"]},
            {"type": "string"},
            {},
            None,
            [],
        ),
        (
            "Invalid Fn::FindInMap too long",
            {"Fn::FindInMap": ["foo", "bar", "key", "key2"]},
            {"type": "string"},
            {},
            None,
            [
                ValidationError(
                    "['foo', 'bar', 'key', 'key2'] is too long (3)",
                    path=deque(["Fn::FindInMap"]),
                    schema_path=deque(["maxItems"]),
                    validator="fn_findinmap",
                ),
            ],
        ),
        (
            "Invalid Fn::FindInMap with wrong type",
            {"Fn::FindInMap": {"foo": "bar"}},
            {"type": "string"},
            {},
            None,
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'array'",
                    path=deque(["Fn::FindInMap"]),
                    schema_path=deque(["type"]),
                    validator="fn_findinmap",
                ),
            ],
        ),
        (
            "Invalid Fn::FindInMap with wrong function",
            {"Fn::FindInMap": [{"Fn::GetAtt": "MyResource.Arn"}, "foo", "bar"]},
            {"type": "string"},
            {},
            None,
            [
                ValidationError(
                    "{'Fn::GetAtt': 'MyResource.Arn'} is not of type 'string'",
                    path=deque(["Fn::FindInMap", 0]),
                    schema_path=deque(["fn_items", "type"]),
                    validator="fn_findinmap",
                ),
            ],
        ),
        (
            "Valid Fn::FindInMap",
            {"Fn::FindInMap": ["A", "B", "C", {"DefaultValue": "D"}]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            None,
            [],
        ),
        (
            "Invalid Fn::FindInMap options not of type object",
            {"Fn::FindInMap": ["A", "B", "C", []]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            None,
            [
                ValidationError(
                    "[] is not of type 'object'",
                    path=deque(["Fn::FindInMap", 3]),
                    schema_path=deque(["fn_items", "type"]),
                    validator="fn_findinmap",
                ),
            ],
        ),
        (
            "Invalid Fn::FindInMap default keyword doesn't exist",
            {"Fn::FindInMap": ["A", "B", "C", {}]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            None,
            [
                ValidationError(
                    "'DefaultValue' is a required property",
                    path=deque(["Fn::FindInMap", 3]),
                    schema_path=deque(["fn_items", "required"]),
                    validator="fn_findinmap",
                ),
            ],
        ),
        (
            "Invalid Fn::FindInMap with a Ref to a resource",
            {"Fn::FindInMap": ["A", {"Ref": "MyResource"}, "C"]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [ValidationError("Foo")],
            [
                ValidationError(
                    "Foo",
                    path=deque(["Fn::FindInMap", 1]),
                    schema_path=deque(["fn_items", "ref"]),
                    validator="ref",
                ),
            ],
        ),
        (
            "Invalid Fn::FindInMap with a bad map key",
            {"Fn::FindInMap": ["A", "C", "B"]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [ValidationError("Foo")],
            [
                ValidationError(
                    "'C' is not one of ['B'] for mapping 'A'",
                    path=deque(["Fn::FindInMap", 1]),
                    schema_path=deque([]),
                ),
            ],
        ),
        (
            "Valid Fn::FindInMap as the Ref could work",
            {"Fn::FindInMap": ["A", {"Ref": "MyParameter"}, "C"]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [],
            [],
        ),
        (
            "Valid Fn::FindInMap with a Ref to AWS::NoValue",
            {
                "Fn::FindInMap": [
                    "A",
                    "B",
                    "C",
                    {"DefaultValue": {"Ref": "AWS::NoValue"}},
                ]
            },
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [],
            [],
        ),
    ],
)
def test_validate(
    name,
    instance,
    schema,
    ref_mock_values,
    context_evolve,
    expected,
    rule,
    context,
    cfn,
):
    context = context.evolve(**context_evolve)
    ref_mock = MagicMock()
    ref_mock.return_value = iter(ref_mock_values or [])
    validator = CfnTemplateValidator({}).extend(validators={"ref": ref_mock})(
        context=context, cfn=cfn
    )
    errs = list(rule.fn_findinmap(validator, schema, instance, {}))

    if ref_mock_values is None:
        ref_mock.assert_not_called()
    else:
        assert ref_mock.call_count == len(ref_mock_values) or 1

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.extensions.aws_gamelift_fleet
from cfnlint.rules.jsonschema.CfnLintJsonSchema import SchemaDetails
from cfnlint.rules.jsonschema.CfnLintJsonSchemaRegional import CfnLintJsonSchemaRegional


class FleetEc2InstanceTypeEnum(CfnLintJsonSchemaRegional):
    id = "E3641"
    shortdesc = "Validate GameLift Fleet EC2 instance type"
    description = (
        "Validates the GameLift Fleet EC2 instance types based on "
        "region and data gathered from the pricing APIs"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::GameLift::Fleet/Properties/EC2InstanceType"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_gamelift_fleet,
                filename="ec2instancetype_enum.json",
            ),
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.cloudwatch.AlarmPeriod import AlarmPeriod


@pytest.fixture(scope="module")
def rule():
    rule = AlarmPeriod()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "Period": 60,
            },
            [],
        ),
        (
            [],  # wrong type
            [],
        ),
        (
            {
                "Period": {"Ref": "Period"},  # ignore when object
            },
            [],
        ),
        (
            {
                "Period": 30,
            },
            [],
        ),
        (
            {
                "Period": 600,
            },
            [],
        ),
        (
            {
                "Period": 45,
            },
            [
                ValidationError(
                    "45 is not one of [10, 30, 60] or a multiple of 60",
                    rule=AlarmPeriod(),
                    path=deque(["Period"]),
                    validator="enum",
                    schema_path=deque(["then", "properties", "Period", "then", "enum"]),
                )
            ],
        ),
        (
            {
                "Period": "45",
            },
            [
                ValidationError(
                    "'45' is not one of [10, 30, 60] or a multiple of 60",
                    rule=AlarmPeriod(),
                    path=deque(["Period"]),
                    validator="enum",
                    schema_path=deque(["then", "properties", "Period", "then", "enum"]),
                )
            ],
        ),
        (
            {
                "Period": 121,
            },
            [
                ValidationError(
                    "121 is not one of [10, 30, 60] or a multiple of 60",
                    rule=AlarmPeriod(),
                    path=deque(["Period"]),
                    validator="multipleOf",
                    schema_path=deque(
                        ["then", "properties", "Period", "else", "multipleOf"]
                    ),
                )
            ],
        ),
        (
            {
                "Period": "121",
            },
            [
                ValidationError(
                    "'121' is not one of [10, 30, 60] or a multiple of 60",
                    rule=AlarmPeriod(),
                    path=deque(["Period"]),
                    validator="multipleOf",
                    schema_path=deque(
                        ["then", "properties", "Period", "else", "multipleOf"]
                    ),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import logging
from test.testlib.testcase import BaseTestCase
from unittest.mock import patch

import yaml
from yaml.scanner import ScannerError

import cfnlint.decode.cfn_json
import cfnlint.decode.cfn_yaml
import cfnlint.decode.decode


class TestDecode(BaseTestCase):
    def tearDown(self) -> None:
        super().tearDown()
        logger = logging.getLogger("cfnlint.decode.decode")
        logger.disabled = False

    def setUp(self):
        """Setup"""
        logger = logging.getLogger("cfnlint.decode.decode")
        logger.disabled = True

    @patch("cfnlint.decode.cfn_yaml.load")
    def test_decode_permission_error(self, mock_cfn_yaml):
        mock_cfn_yaml.side_effect = OSError(13, "Permission denied")
        _, matches = cfnlint.decode.decode(filename="test")
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].rule.id, "E0000")

    @patch("cfnlint.decode.cfn_yaml.load")
    def test_decode_unicode_error(self, mock_cfn_yaml):
        mock_cfn_yaml.side_effect = UnicodeDecodeError("", b"\x00\x00", 0, 0, "reason")
        _, matches = cfnlint.decode.decode(filename="test")
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].rule.id, "E0000")

    @patch("cfnlint.decode.cfn_yaml.load")
    def test_decode_scanner_error_other(self, mock_cfn_yaml):
        mock_cfn_yaml.side_effect = ScannerError(
            problem="foo", problem_mark=yaml.Mark("test", 0, 0, 0, "", 0)
        )
        _, matches = cfnlint.decode.decode(filename="test")
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].rule.id, "E0000")
        self.assertEqual(matches[0].message, "foo")

    @patch("cfnlint.decode.cfn_yaml.load")
    @patch("cfnlint.decode.cfn_json.load")
    def test_decode_scanner_error_tab(self, mock_cfn_json, mock_cfn_yaml):
        mock_cfn_json.return_value = {}
        mock_cfn_yaml.side_effect = ScannerError(
            problem="found character '\\t' that cannot start any token"
        )
        _, matches = cfnlint.decode.decode("test")
        self.assertListEqual(matches, [])
        mock_cfn_json.assert_called_once()

    @patch("cfnlint.decode.cfn_yaml.load")
    @patch("cfnlint.decode.cfn_json.load")
    def test_decode_json_decoder_no_json_error(self, mock_cfn_json, mock_cfn_yaml):
        mock_cfn_json.side_effect = json.JSONDecodeError(
            "No JSON object could be decoded", "", 0
        )
        mock_cfn_yaml.side_effect = ScannerError(
            problem="found character '\\t' that cannot start any token",
            problem_mark=yaml.Mark("test", 0, 0, 0, "", 0),
        )
        _, matches = cfnlint.decode.decode("test")
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].rule.id, "E0000")

    @patch("cfnlint.decode.cfn_yaml.load")
    @patch("cfnlint.decode.cfn_json.load")
    def test_decode_json_decoder_exception_error(self, mock_cfn_json, mock_cfn_yaml):
        mock_cfn_json.side_effect = Exception("foo")
        mock_cfn_yaml.side_effect = ScannerError(
            problem="found character '\\t' that cannot start any token",
            problem_mark=yaml.Mark("test", 0, 0, 0, "", 0),
        )
        _, matches = cfnlint.decode.decode("test")
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].rule.id, "E0000")
        self.assertEqual(
            matches[0].message, "Tried to parse test as JSON but got error: foo"
        )

    @patch("cfnlint.decode.cfn_yaml.load")
    @patch("cfnlint.decode.cfn_json.load")
    def test_decode_json_decoder_expecting_value_error(
        self, mock_cfn_json, mock_cfn_yaml
    ):
        mock_cfn_json.side_effect = json.JSONDecodeError("Expecting value", "", 0)
        mock_cfn_yaml.side_effect = ScannerError(
            problem="found character '\\t' that cannot start any token",
            problem_mark=yaml.Mark("test", 0, 99, 0, "", 0),
        )
        _, matches = cfnlint.decode.decode("test")
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].rule.id, "E0000")
        self.assertEqual(matches[0].linenumber, 100)

    @patch("cfnlint.decode.cfn_yaml.load")
    @patch("cfnlint.decode.cfn_json.load")
    def test_decode_json_decoder_error(self, mock_cfn_json, mock_cfn_yaml):
        err_msg = "Foo"
        mock_cfn_json.side_effect = json.JSONDecodeError(err_msg, "", 0)
        mock_cfn_yaml.side_effect = ScannerError(
            problem="found character '\\t' that cannot start any token"
        )
        _, matches = cfnlint.decode.decode("test")
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].rule.id, "E0000")
        self.assertEqual(matches[0].message, "Foo")

    @patch("cfnlint.decode.cfn_yaml.load")
    def test_decode_yaml_error(self, mock_cfn_yaml):
        err_msg = "Foo"
        mock_cfn_yaml.side_effect = yaml.YAMLError(err_msg)
        _, matches = cfnlint.decode.decode("test")
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].rule.id, "E0000")
        self.assertEqual(matches[0].message, err_msg)

    def test_decode_yaml_null_key(self):
        err_msg = "Null key 'null' not supported (line 3)"
        with self.assertRaises(cfnlint.decode.cfn_yaml.CfnParseError) as e:
            cfnlint.decode.cfn_yaml.loads(
                """
                Parameters:
                    null: test
            """
            )
        self.assertEqual(str(e.exception), err_msg)

    def test_decode_yaml_empty(self):
        template = cfnlint.decode.cfn_yaml.loads("")
        self.assertEqual(template, {})



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword
from cfnlint.schema.manager import PROVIDER_SCHEMA_MANAGER


class ResourceType(CfnLintKeyword):
    """Check Base Resource Configuration"""

    id = "E3006"
    shortdesc = "Validate the CloudFormation resource type"
    description = "Resource types are validated against the spec accounting for regions"
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/*",
            ],
        )

    def validate(self, validator: Validator, keywords: Any, instance: Any, schema: Any):
        resource_type = instance.get("Type")
        if not validator.is_type(resource_type, "string"):
            return

        resource_condition = instance.get("Condition")

        for region in validator.context.regions:
            if validator.is_type(resource_condition, "string"):
                if validator.cfn is None:
                    continue
                if False in validator.cfn.conditions.build_scenerios_on_region(
                    resource_condition, region
                ):
                    continue
            if resource_type in PROVIDER_SCHEMA_MANAGER.get_resource_types(
                region=region
            ):
                continue
            if not resource_type.startswith(
                ("Custom::", "AWS::Serverless::")
            ) and not resource_type.endswith("::MODULE"):
                yield ValidationError(
                    f"Resource type {resource_type!r} does not exist in {region!r}",
                    path=deque(["Type"]),
                    rule=self,
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import os
from test.testlib.testcase import BaseTestCase
from unittest.mock import patch

from cfnlint.decode import cfn_yaml, convert_dict
from cfnlint.schema.manager import ProviderSchemaManager
from cfnlint.template import Template  # pylint: disable=E0401


class TestTemplate(BaseTestCase):
    """Test Template Class in cfnlint"""

    def setUp(self):
        """SetUp template object"""
        filename = "test/fixtures/templates/good/generic.yaml"
        template = self.load_template(filename)
        self.template = Template(filename, template)
        self.resource_names = [
            "IamPipeline",
            "RootInstanceProfile",
            "RolePolicies",
            "MyEC2Instance",
            "RootRole",
            "mySnsTopic",
            "MyEC2Instance1",
            "ElasticLoadBalancer",
            "MyModule",
        ]
        self.parameter_names = ["WebServerPort", "Package", "Package1", "pIops"]

    def test_build_graph(self):
        self.template.build_graph()

        dot = "test/fixtures/templates/good/generic.yaml.dot"

        # ruff: noqa: E501
        expected_content = """digraph "template" {
MyModule [label="MyModule\\n<My::Organization::Custom::MODULE>", color=black, shape=ellipse, type=Resource];
RootRole [label="RootRole\\n<AWS::IAM::Role>", color=black, shape=ellipse, type=Resource];
RolePolicies [label="RolePolicies\\n<AWS::IAM::Policy>", color=black, shape=ellipse, type=Resource];
RootInstanceProfile [label="RootInstanceProfile\\n<AWS::IAM::InstanceProfile>", color=black, shape=ellipse, type=Resource];
MyEC2Instance [label="MyEC2Instance\\n<AWS::EC2::Instance>", color=black, shape=ellipse, type=Resource];
mySnsTopic [label="mySnsTopic\\n<AWS::SNS::Topic>", color=black, shape=ellipse, type=Resource];
MyEC2Instance1 [label="MyEC2Instance1\\n<AWS::EC2::Instance>", color=black, shape=ellipse, type=Resource];
ElasticIP [label="ElasticIP\\n<AWS::EC2::EIP>", color=black, shape=ellipse, type=Resource];
ElasticLoadBalancer [label="ElasticLoadBalancer\\n<AWS::ElasticLoadBalancing::LoadBalancer>", color=black, shape=ellipse, type=Resource];
IamPipeline [label="IamPipeline\\n<AWS::CloudFormation::Stack>", color=black, shape=ellipse, type=Resource];
CustomResource [label="CustomResource\\n<Custom::Function>", color=black, shape=ellipse, type=Resource];
WaitCondition [label="WaitCondition\\n<AWS::CloudFormation::WaitCondition>", color=black, shape=ellipse, type=Resource];
LambdaFunction [label="LambdaFunction\\n<AWS::Lambda::Function>", color=black, shape=ellipse, type=Resource];
RolePolicies -> RootRole [key=0, source_paths="['Properties', 'Roles', 0]", label=Ref, color=black];
RootInstanceProfile -> RootRole [key=0, source_paths="['Properties', 'Roles', 0]", label=Ref, color=black];
MyEC2Instance -> RootInstanceProfile [key=0, source_paths="['Properties', 'IamInstanceProfile']", label=Ref, color=black];
ElasticLoadBalancer -> MyEC2Instance [key=0, source_paths="['Properties', 'Instances', 0]", label=Ref, color=black];
}
""".split(
            "\n"
        )

        assert os.path.exists(dot)
        with open(dot, "r") as file:
            file_contents = file.read().split("\n")
            # doing set equality instead of string equality because python 2.7 and 3.8 produce the same graph but with different edge order
            assert len(file_contents) == len(
                expected_content
            ), f"Length {len(file_contents)} != {len(expected_content)}"
            assert sorted(file_contents) == sorted(expected_content)

        os.remove(dot)

    def test_get_resources_success(self):
        """Test Success on Get Resources"""
        valid_resource_count = 13
        resources = self.template.get_resources()
        assert (
            len(resources) == valid_resource_count
        ), "Expected {} resources, got {}".format(valid_resource_count, len(resources))

    def test_get_modules_success(self):
        """Test Success on Get Modules"""
        expected_modules_count = 1
        actual_modules = self.template.get_modules()
        assert (
            len(actual_modules) == expected_modules_count
        ), "Expected {} modules, got {}".format(
            expected_modules_count, len(actual_modules)
        )

    def test_get_resource_children(self):
        """Test Success on Get Resources"""
        resources = list(self.template.get_resource_children("RootRole"))
        self.assertListEqual(
            resources, ["RolePolicies", "RootInstanceProfile"]
        ), "Expected {} resources, got {}".format(
            ["RolePolicies", "RootInstanceProfile"], resources
        )

        resources = list(
            self.template.get_resource_children(
                "RootRole", ["AWS::IAM::InstanceProfile"]
            )
        )
        self.assertListEqual(
            resources, ["RootInstanceProfile"]
        ), "Expected {} resources, got {}".format(["RootInstanceProfile"], resources)

        resources = list(self.template.get_resource_children("ElasticIP", []))
        self.assertListEqual(resources, []), "Expected {} resources, got {}".format(
            [], resources
        )

    def test_get_resources_bad(self):
        """Don't get resources that aren't properly configured"""
        template = {
            "Resources": {
                "Properties": {"BucketName": "bucket_test"},
                "Type": "AWS::S3::Bucket",
            }
        }
        self.template = Template("test.yaml", template)
        resources = self.template.get_resources()
        assert resources == {}

    def test_get_valid_refs(self):
        """Get Valid REFs"""
        valid_ref_count = 28
        refs = self.template.get_valid_refs()
        assert len(refs) == valid_ref_count, "Expected {} refs, got {}".format(
            valid_ref_count, len(refs)
        )

    def test_conditions_return_object_success(self):
        """Test condition object response and nested IFs"""
        template = [
            "isProd",
            {"Key": "Environment1", "Value": "Prod"},
            {
                "Fn::If": [
                    "isDev",
                    {"Key": "Environment2", "Value": "Dev"},
                    {"Ref": "AWS::NoValue"},
                ]
            },
        ]

        results = self.template.get_condition_values(template, [])
        self.assertEqual(
            results,
            [
                {"Path": [1], "Value": {"Value": "Prod", "Key": "Environment1"}},
                {
                    "Path": [2, "Fn::If", 1],
                    "Value": {"Value": "Dev", "Key": "Environment2"},
                },
            ],
        )

    def test_conditions_return_list_success(self):
        """Test condition list response"""
        template = ["PrimaryRegion", ["EN"], ["BE", "LU", "NL"]]

        results = self.template.get_condition_values(template, [])
        self.assertEqual(
            results,
            [
                {"Value": ["EN"], "Path": [1]},
                {"Value": ["BE", "LU", "NL"], "Path": [2]},
            ],
        )

    def test_conditions_return_string_success(self):
        """Test condition object response and nested IFs"""
        template = ["isProd", {"Ref": "Sample"}, "String"]

        results = self.template.get_condition_values(template, [])
        self.assertEqual(
            results,
            [
                {"Path": [1], "Value": {"Ref": "Sample"}},
                {"Path": [2], "Value": "String"},
            ],
        )

    def test_is_resource_available(self):
        """Test is resource available"""
        temp_obj = convert_dict(
            {
                "Mappings": {"location": {"us-east-1": {"primary": "True"}}},
                "Conditions": {
                    "isPrimary": {
                        "Fn::Equals": [
                            "True",
                            {
                                "Fn::FindInMap": [
                                    "location",
                                    {"Ref": "AWS::Region"},
                                    "primary",
                                ]
                            },
                        ]
                    },
                    "isNotPrimary": {"Fn::Not": [{"Condition": "isPrimary"}]},
                },
                "Resources": {
                    "LambdaExecutionRole": {
                        "Condition": "isPrimary",
                        "Type": "AWS::IAM::Role",
                        "Properties": {},
                    },
                    "AMIIDLookup": {
                        "Type": "AWS::Lambda::Function",
                        "Properties": {
                            "Handler": "index.handler",
                            "Role": {
                                "Fn::If": [
                                    "isPrimary",
                                    {"Fn::GetAtt": "LambdaExecutionRole.Arn"},
                                    {"Ref": "AWS::NoValue"},
                                ]
                            },
                        },
                    },
                    "AMIIDLookupBackup": {
                        "Type": "AWS::Lambda::Function",
                        "Properties": {
                            "Handler": "index.handler",
                            "Role": {
                                "Fn::If": [
                                    "isPrimary",
                                    {
                                        "Fn::If": [
                                            "isPrimary",
                                            {"Ref": "AWS::NoValue"},
                                            {"Fn::GetAtt": "LambdaExecutionRole.Arn"},
                                        ]
                                    },
                                    {"Fn::GetAtt": "LambdaExecutionRole.Arn"},
                                ]
                            },
                        },
                    },
                },
                "Outputs": {
                    "lambdaArn": {
                        "Condition": "isPrimary",
                        "Value": {"Fn::GetAtt": "LambdaExecutionRole.Arn"},
                    },
                    "AMIIDLookupBackup": {
                        "Value": {
                            "Fn::If": [
                                "isNotPrimary",
                                {
                                    "Fn::If": [
                                        "isNotPrimary",
                                        {"Ref": "AWS::NoValue"},
                                        {"Fn::GetAtt": "LambdaExecutionRole.Arn"},
                                    ]
                                },
                                {"Fn::GetAtt": "LambdaExecutionRole.Arn"},
                            ]
                        }
                    },
                },
            }
        )
        template = Template("test.yaml", temp_obj)
        # Doesn't fail with a Fn::If based condition
        self.assertEqual(
            template.is_resource_available(
                [
                    "Resources",
                    "AMIIDLookup",
                    "Properties",
                    "Role",
                    "Fn::If",
                    1,
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
                "LambdaExecutionRole",
            ),
            [],
        )
        # Failure when the opposite of the resource condition is used
        self.assertEqual(
            template.is_resource_available(
                [
                    "Resources",
                    "AMIIDLookupBackup",
                    "Properties",
                    "Role",
                    "Fn::If",
                    2,
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
                "LambdaExecutionRole",
            ),
            [{"isPrimary": False}],
        )
        # No error when the condition has both true/false use cases
        self.assertEqual(
            template.is_resource_available(
                [
                    "Outputs",
                    "AMIIDLookupBackup",
                    "Value",
                    "Fn::If",
                    1,
                    "Fn::If",
                    2,
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
                "LambdaExecutionRole",
            ),
            [],
        )
        # Doesn't fail when the Output has a Condition defined
        self.assertEqual(
            template.is_resource_available(
                [
                    "Outputs",
                    "lambdaArn",
                    "Value",
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
                "LambdaExecutionRole",
            ),
            [],
        )
        # Doesn't fail when the Resource doesn't exist
        self.assertEqual(
            template.is_resource_available(
                [
                    "Outputs",
                    "lambdaArn",
                    "Value",
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
                "UnknownResource",
            ),
            [],
        )

    def test_is_resource_not_available(self):
        """Test is resource available"""
        temp_obj = convert_dict(
            {
                "Mappings": {"location": {"us-east-1": {"primary": "True"}}},
                "Conditions": {
                    "isPrimary": {
                        "Fn::Equals": [
                            "True",
                            {
                                "Fn::FindInMap": [
                                    "location",
                                    {"Ref": "AWS::Region"},
                                    "primary",
                                ]
                            },
                        ]
                    },
                },
                "Resources": {
                    "LambdaExecutionRole": {
                        "Condition": "isPrimary",
                        "Type": "AWS::IAM::Role",
                        "Properties": {},
                    },
                    "AMIIDLookup": {
                        "Type": "AWS::Lambda::Function",
                        "Properties": {
                            "Handler": "index.handler",
                            "Role": {"Fn::GetAtt": "LambdaExecutionRole.Arn"},
                        },
                    },
                },
                "Outputs": {
                    "lambdaArn": {"Value": {"Fn::GetAtt": "LambdaExecutionRole.Arn"}}
                },
            }
        )
        template = Template("test.yaml", temp_obj)
        self.assertEqual(
            template.is_resource_available(
                [
                    "Resources",
                    "AMIIDLookup",
                    "Properties",
                    "Role",
                    "Fn::If",
                    1,
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
                "LambdaExecutionRole",
            ),
            [{"isPrimary": False}],
        )
        self.assertEqual(
            template.is_resource_available(
                [
                    "Outputs",
                    "lambdaArn",
                    "Value",
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
                "LambdaExecutionRole",
            ),
            [{"isPrimary": False}],
        )
        # Doesn't fail when the Path doesn't exist.  Still shows an error based on the fact there
        # are no conditions in the path that is workable
        self.assertEqual(
            template.is_resource_available(
                [
                    "Resources",
                    "AMIIDLookup",
                    "Properties",
                    "BadProperty",
                    "Fn::If",
                    1,
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
                "LambdaExecutionRole",
            ),
            [{"isPrimary": False}],
        )
        # Gives appropriate response when there is no condition in path
        self.assertEqual(
            template.is_resource_available(
                ["Resources", "AMIIDLookup", "Properties", "Handler"],
                "LambdaExecutionRole",
            ),
            [{"isPrimary": False}],
        )

    def test_get_conditions_from_path(self):
        """Test is resource available"""
        temp_obj = convert_dict(
            {
                "Resources": {
                    "AMIIDLookup": {
                        "Type": "AWS::Lambda::Function",
                        "Properties": {
                            "Role": {
                                "Fn::If": [
                                    "isPrimary",
                                    {"Fn::GetAtt": "LambdaExecutionRole.Arn"},
                                    {"Ref": "AWS::NoValue"},
                                ]
                            }
                        },
                    },
                    "InstanceProfile": {
                        "Condition": "isPrimaryAndProduction",
                        "Type": "AWS::IAM::InstanceProfile",
                        "Properties": {
                            "Path": "/",
                            "Roles": [{"Ref": "LambdaExecutionRole"}],
                        },
                    },
                },
                "Outputs": {
                    "lambdaArn": {
                        "Condition": "isPrimary",
                        "Value": {"Fn::GetAtt": "LambdaExecutionRole.Arn"},
                    }
                },
            }
        )
        template = Template("test.yaml", temp_obj)
        # Gets the condition when in the middle of the Path
        self.assertEqual(
            template.get_conditions_from_path(
                template.template,
                [
                    "Resources",
                    "AMIIDLookup",
                    "Properties",
                    "Role",
                    "Fn::If",
                    1,
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
            ),
            {"isPrimary": {True}},
        )
        self.assertEqual(
            template.get_conditions_from_path(
                template.template,
                ["Resources", "AMIIDLookup", "Properties", "Role", "Fn::If"],
            ),
            {"isPrimary": {True, False}},
        )
        self.assertEqual(
            template.get_conditions_from_path(
                template.template,
                [
                    "Outputs",
                    "lambdaArn",
                    "Value",
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
            ),
            {"isPrimary": {True}},
        )
        self.assertEqual(
            template.get_conditions_from_path(
                template.template,
                [
                    "Resources",
                    "InstanceProfile",
                    "Properties",
                    "Roles",
                    0,
                    "Ref",
                    "LambdaExecutionRole",
                ],
            ),
            {"isPrimaryAndProduction": {True}},
        )
        self.assertEqual(
            template.get_conditions_from_path(
                template.template,
                [
                    "Resources",
                    "AMIIDLookup",
                    "Properties",
                    "Handler",
                    "Ref",
                    "LambdaHandler",
                ],
            ),
            {},
        )

    def test_failure_get_conditions_from_path(self):
        """Test get conditions from path when things arne't formatted correctly"""
        temp_obj = convert_dict(
            {
                "Resources": {
                    "AMIIDLookup": {
                        "Type": "AWS::Lambda::Function",
                        "Properties": {
                            "Role": {
                                "Fn::If": [
                                    "isPrimary",
                                    {"Fn::GetAtt": "LambdaExecutionRole.Arn"},
                                ]
                            }
                        },
                    },
                    "myInstance4": {
                        "Type": "AWS::EC2::Instance",
                        "Properties": {
                            "InstanceType": {
                                "Fn::If": {
                                    "Fn::If": ["isPrimary", "t3.2xlarge", "t3.xlarge"]
                                }
                            }
                        },
                    },
                }
            }
        )
        template = Template("test.yaml", temp_obj)
        # Gets no condition names when it isn't a valid list
        self.assertEqual(
            template.get_conditions_from_path(
                template.template,
                [
                    "Resources",
                    "AMIIDLookup",
                    "Properties",
                    "Role",
                    "Fn::If",
                    1,
                    "Fn::GetAtt",
                    ["LambdaExecutionRole", "Arn"],
                ],
            ),
            {},
        )
        # Gets no condition names when it isn't really a list
        self.assertEqual(
            template.get_conditions_from_path(
                template.template,
                [
                    "Resources",
                    "myInstance4",
                    "Properties",
                    "InstanceType",
                    "Fn::If",
                    "Fn::If",
                    0,
                ],
            ),
            {},
        )

    def test_get_conditions_scenarios_from_object(self):
        """Test Getting condition names in an object/list"""
        template = {
            "Conditions": {
                "isProduction": {"Fn::Equals": [{"Ref": "myEnvironment"}, "Prod"]},
                "isPrimary": {
                    "Fn::Equals": [
                        "True",
                        {
                            "Fn::FindInMap": [
                                "location",
                                {"Ref": "AWS::Region"},
                                "primary",
                            ]
                        },
                    ]
                },
                "isPrimaryAndProduction": {
                    "Fn::And": [
                        {"Condition": "isProduction"},
                        {"Condition": "isPrimary"},
                    ]
                },
                "isDevelopment": {"Fn::Equals": [{"Ref": "myEnvironment"}, "Dev"]},
            },
            "Resources": {
                "myInstance1": {
                    "Type": "AWS::EC2::Instance",
                    "Properties": {
                        "Fn::If": [
                            "isPrimaryAndProduction",
                            {"ImageId": "ami-1234567", "InstanceType": "t2.medium"},
                            {
                                "Fn::If": [
                                    "isDevelopment",
                                    {
                                        "ImageId": "ami-abcdefg",
                                        "InstanceType": "t2.xlarge",
                                    },
                                    {
                                        "ImageId": "ami-123abcd",
                                        "InstanceType": "m3.medium",
                                    },
                                ]
                            },
                        ]
                    },
                },
                "myInstance2": {
                    "Type": "AWS::EC2::Instance",
                    "Properties": {
                        "ImageId": {
                            "Fn::If": ["isProduction", "ami-123456", "ami-abcdef"]
                        },
                        "InstanceType": {
                            "Fn::If": [
                                "isPrimaryAndProduction",
                                {"Fn::If": ["isPrimary", "t3.2xlarge", "t3.xlarge"]},
                                "t3.medium",
                            ]
                        },
                    },
                },
            },
        }
        template = Template("test.yaml", template)
        # includes nested IFs
        self.assertEqualListOfDicts(
            template.get_conditions_scenarios_from_object(
                template.template.get("Resources", {})
                .get("myInstance2", {})
                .get("Properties", {})
            ),
            [
                {
                    "isPrimary": True,
                    "isProduction": True,
                    "isPrimaryAndProduction": True,
                },
                {
                    "isPrimary": True,
                    "isProduction": False,
                    "isPrimaryAndProduction": False,
                },
                {
                    "isPrimary": False,
                    "isProduction": True,
                    "isPrimaryAndProduction": False,
                },
                {
                    "isPrimary": False,
                    "isProduction": False,
                    "isPrimaryAndProduction": False,
                },
            ],
        )
        self.assertEqualListOfDicts(
            template.get_conditions_scenarios_from_object(
                template.template.get("Resources", {})
                .get("myInstance1", {})
                .get("Properties", {})
            ),
            [
                {"isPrimaryAndProduction": True, "isDevelopment": False},
                {"isPrimaryAndProduction": False, "isDevelopment": True},
                {"isPrimaryAndProduction": False, "isDevelopment": False},
            ],
        )

    def test_get_object_without_conditions(self):
        """Test Getting condition names in an object/list"""
        template = convert_dict(
            {
                "Conditions": {
                    "useAmiId": {"Fn::Not": [{"Fn::Equals": [{"Ref": "myAmiId"}, ""]}]}
                },
                "Resources": {
                    "myInstance": {
                        "Type": "AWS::EC2::Instance",
                        "Properties": {
                            "ImageId": {
                                "Fn::If": [
                                    "useAmiId",
                                    {"Ref": "myAmiId"},
                                    {"Ref": "AWS::NoValue"},
                                ]
                            },
                            "LaunchConfiguration": {
                                "Fn::If": [
                                    "useAmiId",
                                    {"Ref": "AWS::NoValue"},
                                    {"Ref": "LaunchConfiguration"},
                                ]
                            },
                        },
                    }
                },
            }
        )
        template = Template("test.yaml", template)

        results = template.get_object_without_conditions(
            template.template.get("Resources", {})
            .get("myInstance", {})
            .get("Properties", {})
        )

        for result in results:
            if result["Scenario"] == {"useAmiId": True}:
                self.assertDictEqual(result["Object"], {"ImageId": {"Ref": "myAmiId"}})
            elif result["Scenario"] == {"useAmiId": False}:
                self.assertDictEqual(
                    result["Object"],
                    {"LaunchConfiguration": {"Ref": "LaunchConfiguration"}},
                )

    def test_get_object_without_conditions_no_value(self):
        """Test Getting condition names in an object/list"""
        template = convert_dict(
            {
                "Conditions": {
                    "CreateAppVolume": {"Fn::Equals": [{"Ref": "myAppVolume"}, "true"]}
                },
                "Resources": {
                    "myInstance": {
                        "Type": "AWS::EC2::Instance",
                        "Properties": {
                            "ImageId": "ami-123456",
                            "BlockDeviceMappings": [
                                {
                                    "DeviceName": "/dev/hda1",
                                    "Ebs": {
                                        "DeleteOnTermination": True,
                                        "VolumeType": "gp2",
                                    },
                                },
                                {
                                    "Fn::If": [
                                        "CreateAppVolume",
                                        {
                                            "DeviceName": "/dev/xvdf",
                                            "Ebs": {
                                                "DeleteOnTermination": True,
                                                "VolumeType": "gp2",
                                            },
                                        },
                                        {"Ref": "AWS::NoValue"},
                                    ]
                                },
                            ],
                        },
                    }
                },
            }
        )
        template = Template("test.yaml", template)

        results = template.get_object_without_conditions(
            template.template.get("Resources")
            .get("myInstance")
            .get("Properties")
            .get("BlockDeviceMappings")[1]
            .get("Fn::If")[2]
        )

        self.assertEqual(results, [])

        results = template.get_object_without_conditions(
            template.template.get("Resources")
            .get("myInstance")
            .get("Properties")
            .get("BlockDeviceMappings")
        )
        # when item is a list return empty list
        self.assertEqual(
            results,
            [
                {
                    "Object": [
                        {
                            "DeviceName": "/dev/hda1",
                            "Ebs": {"DeleteOnTermination": True, "VolumeType": "gp2"},
                        },
                        {
                            "DeviceName": "/dev/xvdf",
                            "Ebs": {"DeleteOnTermination": True, "VolumeType": "gp2"},
                        },
                    ],
                    "Scenario": {"CreateAppVolume": True},
                },
                {
                    "Object": [
                        {
                            "DeviceName": "/dev/hda1",
                            "Ebs": {"DeleteOnTermination": True, "VolumeType": "gp2"},
                        }
                    ],
                    "Scenario": {"CreateAppVolume": False},
                },
            ],
        )

        # when item is a string return empty list
        self.assertEqual(
            template.get_object_without_conditions("String"),
            [{"Object": "String", "Scenario": None}],
        )

    def test_get_object_without_conditions_for_list(self):
        """Test Getting condition names in an object/list"""
        template = convert_dict(
            {
                "Conditions": {
                    "CreateAppVolume": {"Fn::Equals": [{"Ref": "CreateVolums"}, "true"]}
                },
                "Resources": {
                    "myInstance": {
                        "Type": "AWS::EC2::Instance",
                        "Properties": {
                            "ImageId": "ami-123456",
                            "BlockDeviceMappings": {
                                "Fn::If": [
                                    "CreateAppVolume",
                                    [
                                        {
                                            "DeviceName": "/dev/hda1",
                                            "Ebs": {
                                                "DeleteOnTermination": True,
                                                "VolumeType": "gp2",
                                            },
                                        },
                                        {
                                            "DeviceName": "/dev/xvdf",
                                            "Ebs": {
                                                "DeleteOnTermination": True,
                                                "VolumeType": "gp2",
                                            },
                                        },
                                    ],
                                    {"Ref": "AWS::NoValue"},
                                ]
                            },
                        },
                    }
                },
            }
        )

        template = Template("test.yaml", template)

        results = template.get_object_without_conditions(
            template.template.get("Resources")
            .get("myInstance")
            .get("Properties")
            .get("BlockDeviceMappings")
        )
        # handles IFs for a list
        self.assertEqual(
            results,
            [
                {
                    "Scenario": {"CreateAppVolume": True},
                    "Object": [
                        {
                            "DeviceName": "/dev/hda1",
                            "Ebs": {"DeleteOnTermination": True, "VolumeType": "gp2"},
                        },
                        {
                            "DeviceName": "/dev/xvdf",
                            "Ebs": {"DeleteOnTermination": True, "VolumeType": "gp2"},
                        },
                    ],
                }
            ],
        )

    def test_get_object_without_going_to_far(self):
        # Validate that we don't go too deep in determining conditions to evaluate
        template = convert_dict(
            {
                "Conditions": {
                    "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
                    "IsUsWest2": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-west-2"]},
                },
                "Resources": {
                    "MyFoo": {
                        "Type": "MyFoo",
                        "Properties": {
                            "Foo": {"Fn::If": ["IsUsEast1", "A", "1"]},
                            "Bar": {
                                "FooBar": {
                                    "Fn::If": [
                                        "IsUsWest2",
                                        True,
                                        False,
                                    ]
                                }
                            },
                        },
                    }
                },
            }
        )

        template = Template("test.yaml", template)

        results = template.get_object_without_conditions(
            template.template.get("Resources").get("MyFoo").get("Properties")
        )

        self.maxDiff = None
        # handles IFs for a list
        self.assertEqual(
            results,
            [
                {
                    "Scenario": {"IsUsEast1": True},
                    "Object": {
                        "Foo": "A",
                        "Bar": {
                            "FooBar": {
                                "Fn::If": [
                                    "IsUsWest2",
                                    True,
                                    False,
                                ]
                            }
                        },
                    },
                },
                {
                    "Scenario": {"IsUsEast1": False},
                    "Object": {
                        "Foo": "1",
                        "Bar": {
                            "FooBar": {
                                "Fn::If": [
                                    "IsUsWest2",
                                    True,
                                    False,
                                ]
                            }
                        },
                    },
                },
            ],
        )

    def test_get_object_with_root_f(self):
        # Validate that we don't go too deep in determining conditions to evaluate
        template = convert_dict(
            {
                "Conditions": {
                    "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
                },
                "Resources": {
                    "MyFoo": {
                        "Type": "MyFoo",
                        "Properties": {
                            "Fn::If": ["IsUsEast1", {"Foo": "Bar"}, {"Bar": "Foo"}]
                        },
                    }
                },
            }
        )

        template = Template("test.yaml", template)

        results = template.get_object_without_conditions(
            template.template.get("Resources").get("MyFoo").get("Properties")
        )

        self.maxDiff = None
        # handles IFs for a list
        self.assertEqual(
            results,
            [
                {
                    "Scenario": {"IsUsEast1": True},
                    "Object": {"Foo": "Bar"},
                },
                {
                    "Scenario": {"IsUsEast1": False},
                    "Object": {"Bar": "Foo"},
                },
            ],
        )

    def test_get_object_without_conditions_for_bad_formats(self):
        """Test Getting condition names in an object/list"""
        template = convert_dict(
            {
                "Conditions": {
                    "CreateAppVolume": {"Fn::Equals": [{"Ref": "CreateVolums"}, "true"]}
                },
                "Resources": {
                    "myInstance": {
                        "Type": "AWS::EC2::Instance",
                        "Properties": {
                            "ImageId": "ami-123456",
                            "BlockDeviceMappings": {
                                "Fn::If": ["CreateAppVolume", {"Ref": "AWS::NoValue"}]
                            },
                        },
                    },
                    "myInstance1": {
                        "Type": "AWS::EC2::Instance",
                        "Properties": {
                            "ImageId": "ami-123456",
                            "BlockDeviceMappings": {"Fn::If": ["CreateAppVolume"]},
                        },
                    },
                },
            }
        )

        template = Template("test.yaml", template)

        results = template.get_object_without_conditions(
            template.template.get("Resources")
            .get("myInstance")
            .get("Properties")
            .get("BlockDeviceMappings")
        )
        # There is no False here but it still passes the 1st item back
        self.assertEqual(
            results,
            [
                {
                    "Object": {"Fn::If": ["CreateAppVolume", {"Ref": "AWS::NoValue"}]},
                    "Scenario": None,
                }
            ],
        )

        results = template.get_object_without_conditions(
            template.template.get("Resources")
            .get("myInstance1")
            .get("Properties")
            .get("BlockDeviceMappings")
        )
        # There is no False here but it still passes the 1st item back
        self.assertEqual(
            results, [{"Object": {"Fn::If": ["CreateAppVolume"]}, "Scenario": None}]
        )

    def test_get_object_without_nested_conditions(self):
        """Test Getting condition names in an object/list"""
        template = convert_dict(
            {
                "Conditions": {
                    "isProduction": {"Fn::Equals": [{"Ref": "myEnvironment"}, "prod"]},
                    "isDevelopment": {"Fn::Equals": [{"Ref": "myEnvironment"}, "dev"]},
                },
                "Resources": {
                    "myInstance": {
                        "Type": "AWS::EC2::Instance",
                        "Properties": {
                            "ImageId": {
                                "Fn::If": [
                                    "isProduction",
                                    "ami-prod1234",
                                    {
                                        "Fn::If": [
                                            "isDevelopment",
                                            "ami-dev1234",
                                            "ami-lab1234",
                                        ]
                                    },
                                ]
                            }
                        },
                    }
                },
            }
        )
        template = Template("test.yaml", template)

        results = template.get_object_without_conditions(
            template.template.get("Resources", {})
            .get("myInstance", {})
            .get("Properties", {})
        )

        self.assertEqual(len(results), 3)
        for result in results:
            if (
                result["Scenario"]["isProduction"]
                and not result["Scenario"]["isDevelopment"]
            ):
                self.assertDictEqual(result["Object"], {"ImageId": "ami-prod1234"})
            elif result["Scenario"] == {"isProduction": False, "isDevelopment": True}:
                self.assertDictEqual(result["Object"], {"ImageId": "ami-dev1234"})
            elif result["Scenario"] == {"isProduction": False, "isDevelopment": False}:
                self.assertDictEqual(result["Object"], {"ImageId": "ami-lab1234"})
            else:
                # Blanket assertion error
                self.assertDictEqual(result["Scenario"], {})

    def test_get_object_without_nested_conditions_iam(self):
        """Test Getting condition names in an object/list"""
        template = {
            "Conditions": {
                "isProduction": {"Fn::Equals": [{"Ref": "myEnvironment"}, "prod"]},
                "isDevelopment": {"Fn::Equals": [{"Ref": "myEnvironment"}, "dev"]},
            },
            "Resources": {
                "Role": {
                    "Type": "AWS::IAM::Role",
                    "Properties": {
                        "AssumeRolePolicyDocument": {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Sid": "",
                                    "Action": "sts:AssumeRole",
                                    "Effect": "Allow",
                                    "Principal": {
                                        "AWS": {
                                            "Fn::If": [
                                                "isProduction",
                                                "123456789012",
                                                "210987654321",
                                            ]
                                        }
                                    },
                                }
                            ],
                        },
                        "ManagedPolicyArns": [
                            "arn:aws:iam::aws:policy/AdministratorAccess"
                        ],
                        "Path": "/",
                    },
                }
            },
        }

        template = Template("test.yaml", template)

        results = template.get_object_without_nested_conditions(
            template.template.get("Resources", {})
            .get("Role", {})
            .get("Properties", {})
            .get("AssumeRolePolicyDocument", {}),
            ["Resources", "Role", "Properties", "AssumeRolePolicyDocument"],
        )

        self.assertEqual(len(results), 2)
        for result in results:
            if not result["Scenario"]["isProduction"]:
                self.assertDictEqual(
                    result["Object"],
                    {
                        "Version": "2012-10-17",
                        "Statement": [
                            {
                                "Sid": "",
                                "Action": "sts:AssumeRole",
                                "Effect": "Allow",
                                "Principal": {"AWS": "210987654321"},
                            }
                        ],
                    },
                )
            if result["Scenario"]["isProduction"]:
                self.assertDictEqual(
                    result["Object"],
                    {
                        "Version": "2012-10-17",
                        "Statement": [
                            {
                                "Sid": "",
                                "Action": "sts:AssumeRole",
                                "Effect": "Allow",
                                "Principal": {"AWS": "123456789012"},
                            }
                        ],
                    },
                )

    def test_get_object_without_nested_conditions_basic(self):
        """Test Getting condition names in an object/list"""
        template = {"Resources": {"Test": "Test"}}

        template = Template("test.yaml", template)

        results = template.get_object_without_nested_conditions(
            template.template.get("Resources", {}).get("Test", {}),
            ["Resources", "Test"],
        )
        self.assertEqual(results, [])

    def test_get_object_without_nested_conditions_ref(self):
        """Test Getting condition names in an object/list"""
        template = {"Resources": {"Test": {"Ref": "AWS::NoValue"}}}

        template = Template("test.yaml", template)

        results = template.get_object_without_nested_conditions(
            template.template.get("Resources", {}).get("Test", {}),
            ["Resources", "Test"],
        )
        self.assertEqual(results, [{"Object": {}, "Scenario": None}])
        results = template.get_object_without_nested_conditions(
            template.template.get("Resources", {}), ["Resources"]
        )
        self.assertEqual(results, [{"Object": {}, "Scenario": None}])

    def test_get_condition_scenarios_below_path(self):
        """Test Getting condition names in an object/list"""
        template = {
            "Parameters": {"Environment": {"Type": "String"}},
            "Conditions": {
                "isPrimaryRegion": {
                    "Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]
                },
                "isProduction": {"Fn::Equals": [{"Ref": "Environment"}, "prod"]},
            },
            "Resources": {
                "Role": {
                    "Type": "AWS::IAM::Role",
                    "Properties": {
                        "AssumeRolePolicyDocument": {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Sid": "",
                                    "Action": "sts:AssumeRole",
                                    "Effect": "Allow",
                                    "Principal": {
                                        "AWS": {
                                            "Fn::If": [
                                                "isProduction",
                                                {
                                                    "Fn::Sub": [
                                                        "{account}",
                                                        {
                                                            "account": {
                                                                "Fn::If": [
                                                                    "isPrimaryRegion",
                                                                    "123456789012",
                                                                    "ABCDEFGHIJKL",
                                                                ]
                                                            }
                                                        },
                                                    ]
                                                },
                                                "210987654321",
                                            ]
                                        }
                                    },
                                }
                            ],
                        },
                        "ManagedPolicyArns": [
                            "arn:aws:iam::aws:policy/AdministratorAccess"
                        ],
                        "Path": "/",
                    },
                }
            },
        }

        template = Template("test.yaml", template)
        results = template.get_condition_scenarios_below_path(
            ["Resources", "Role", "Properties", "AssumeRolePolicyDocument"]
        )
        self.assertEqualListOfDicts(
            results, [{"isProduction": True}, {"isProduction": False}]
        )
        results = template.get_condition_scenarios_below_path(
            ["Resources", "Role", "Properties", "AssumeRolePolicyDocument"], True
        )
        self.assertEqualListOfDicts(
            results,
            [
                {"isPrimaryRegion": False, "isProduction": False},
                {"isPrimaryRegion": False, "isProduction": True},
                {"isPrimaryRegion": True, "isProduction": False},
                {"isPrimaryRegion": True, "isProduction": True},
            ],
        )

    def test_get_directives(self):
        """Test getting directives from a template"""
        template_yaml = {
            "Resources": {
                "Type": "AWS::S3::Bucket",
                "myBucket": {
                    "Properties": {"BucketName": "bucket_test"},
                    "Type": "AWS::S3::Bucket",
                },
                "myBucket1": {
                    "Metadata": {
                        "cfn-lint": {"config": {"ignore_checks": ["E3012", "I1001"]}}
                    },
                    "Properties": {},
                    "Type": "AWS::S3::Bucket",
                },
                "myBucket2": {
                    "Metadata": {
                        "cfn-lint": {
                            "config": {
                                "ignore_checks": [
                                    "E3012",
                                ]
                            }
                        }
                    },
                    "Properties": {},
                    "Type": "AWS::S3::Bucket",
                },
            }
        }

        template = Template(
            "test.yaml",
            cfn_yaml.loads(
                json.dumps(
                    template_yaml, sort_keys=True, indent=4, separators=(",", ": ")
                )
            ),
        )
        directives = template.get_directives()
        expected_result = {
            "E3012": ["myBucket1", "myBucket2"],
            "I1001": ["myBucket1"],
        }
        self.assertDictEqual(directives, expected_result)

    def test_is_cdk_bad_type(self):
        template = {
            "Resources": {
                "CDK": {
                    "Type": ["AWS::CDK::Metadata"],
                    "Properties": {
                        "AssumeRolePolicyDocument": {
                            "Version": "2012-10-17",
                        }
                    },
                }
            },
        }

        template = Template("test.yaml", template)
        self.assertFalse(template.is_cdk_template())

    def test_is_cdk_bad_resources(self):
        template = {
            "Resources": [
                {
                    "CDK": {
                        "Type": ["AWS::CDK::Metadata"],
                        "Properties": {
                            "AssumeRolePolicyDocument": {
                                "Version": "2012-10-17",
                            }
                        },
                    }
                }
            ],
        }

        template = Template("test.yaml", template)
        self.assertFalse(template.is_cdk_template())

    def test_is_cdk_bad_resource_props(self):
        template = {
            "Resources": {
                "CDK": [
                    {
                        "Type": ["AWS::CDK::Metadata"],
                        "Properties": {
                            "AssumeRolePolicyDocument": {
                                "Version": "2012-10-17",
                            }
                        },
                    }
                ]
            },
        }

        template = Template("test.yaml", template)
        self.assertFalse(template.is_cdk_template())

    def test_schemas_get_att(self):
        """Validate getAtt when using a registry schema"""
        schema_filename = "test/fixtures/registry/custom/"

        filename = "test/fixtures/templates/good/schema_resource.yaml"
        template = self.load_template(filename)
        schema_manager = ProviderSchemaManager()
        schema_manager.load_registry_schemas(schema_filename)
        with patch("cfnlint.template.getatts.PROVIDER_SCHEMA_MANAGER", schema_manager):
            self.template = Template(filename, template)
            self.assertDictEqual(
                {
                    "oneOf": [
                        {
                            "type": "array",
                            "items": [
                                {"type": "string", "enum": ["MyReport"]},
                                {"type": ["string", "object"]},
                            ],
                            "allOf": [
                                {
                                    "if": {
                                        "items": [
                                            {"type": "string", "const": "MyReport"},
                                            {"type": ["string", "object"]},
                                        ]
                                    },
                                    "then": {
                                        "if": {
                                            "items": [
                                                {"type": "string", "const": "MyReport"},
                                                {"type": "string"},
                                            ]
                                        },
                                        "then": {
                                            "items": [
                                                {"type": "string", "const": "MyReport"},
                                                {
                                                    "type": "string",
                                                    "enum": ["TPSCode", "Authors"],
                                                },
                                            ]
                                        },
                                        "else": {
                                            "items": [
                                                {"type": "string", "const": "MyReport"},
                                                {
                                                    "type": "object",
                                                    "properties": {
                                                        "Ref": {"type": "string"}
                                                    },
                                                    "required": ["Ref"],
                                                    "additionalProperties": False,
                                                },
                                            ]
                                        },
                                    },
                                    "else": {},
                                }
                            ],
                        },
                        {
                            "type": "string",
                            "enum": ["MyReport.TPSCode", "MyReport.Authors"],
                        },
                    ]
                },
                self.template.get_valid_getatts().json_schema("us-east-1"),
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules import CloudFormationLintRule


class PropertyNames(CloudFormationLintRule):
    """Check maxLength values are correct"""

    def __init__(self, approaching_limit_rule: str | None = None) -> None:
        super().__init__()
        self.config["threshold"] = 0.9
        self.approaching_limit_rule = approaching_limit_rule

        self.child_rules = {}
        if self.approaching_limit_rule:
            self.child_rules[self.approaching_limit_rule] = None

    def _max_length(
        self, validator: Validator, mL: int, instance: Any, schema: dict[str, Any]
    ):
        if not validator.is_type(instance, "string"):
            return

        percent = len(instance) / mL
        if percent > 1:
            yield ValidationError(message=f"{instance!r} is longer than {mL}")
            return

        if percent > self.config["threshold"] and self.approaching_limit_rule:
            rule = self.child_rules.get(self.approaching_limit_rule)
            if not rule:
                return

            if hasattr(rule, "maxLength") and callable(getattr(rule, "maxLength")):
                validate = getattr(rule, "maxLength")
                yield from validate(validator, mL, instance, schema)
                return

            yield ValidationError(
                f"{instance!r} is approaching the max length of {mL}",
                rule=rule,
            )

    def propertyNames(
        self,
        validator: Validator,
        propertyNames: Any,
        instance: Any,
        schema: dict[str, Any],
    ) -> ValidationResult:
        if not validator.is_type(instance, "object"):
            return

        v = validator.extend(
            validators={
                "maxLength": self._max_length,
            }
        )({})

        for property in instance:
            yield from v.descend(
                instance=property,
                schema=propertyNames,
                path=property,
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import Validator
from cfnlint.rules.functions._BaseFn import BaseFn


class Cidr(BaseFn):
    """Check if Cidr values are correct"""

    id = "E1024"
    shortdesc = "Cidr validation of parameters"
    description = "Making sure the function CIDR is a list with valid values"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-cidr.html"
    tags = ["functions", "cidr"]

    def __init__(self) -> None:
        super().__init__("Fn::Cidr", ("array",), None)
        self.fn_cidr = self.validate

    def schema(self, validator: Validator, instance: Any) -> dict[str, Any]:
        functions = [
            "Fn::FindInMap",
            "Fn::Select",
            "Ref",
            "Fn::GetAtt",
            "Fn::Sub",
            "Fn::ImportValue",
            "Fn::If",
        ]
        return {
            "type": ["array"],
            "maxItems": 3,
            "minItems": 2,
            "fn_items": [
                {
                    "functions": functions,
                    "schema": {
                        "type": ["string"],
                        "pattern": (
                            "^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.)"
                            "{3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])"
                            "(\\/([0-9]|[1-2][0-9]|3[0-2]))$"
                        ),
                    },
                },
                {
                    "functions": functions,
                    "schema": {
                        "type": ["integer"],
                        "minimum": 1,
                        "maximum": 256,
                    },
                },
                {
                    "functions": functions,
                    "schema": {
                        "type": ["integer"],
                        "minimum": 1,
                        "maximum": 128,
                    },
                },
            ],
        }



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.CreationPolicy import CreationPolicy


@pytest.fixture
def rule():
    return CreationPolicy()


@pytest.fixture
def template():
    return {
        "Resources": {
            "MyInstance": {
                "Type": "AWS::EC2::Instance",
            },
            "MyWaitCondition": {
                "Type": "AWS::CloudFormation::WaitCondition",
            },
            "MyAutoScalingGroup": {
                "Type": "AWS::AutoScaling::AutoScalingGroup",
            },
            "MyAppStreamFleet": {
                "Type": "AWS::AppStream::Fleet",
            },
            "MyLambdaFunction": {
                "Type": "AWS::Lambda::Function",
            },
        }
    }


@pytest.mark.parametrize(
    "name, instance, path, expected",
    [
        (
            "Correct for app stream",
            {"StartFleet": {"Type": True}},
            {
                "path": deque(["Resources", "MyAppStreamFleet", "CreationPolicy"]),
            },
            [],
        ),
        (
            "Bad type for app stream",
            {
                "StartFleet": {
                    "Type": {},
                }
            },
            {
                "path": deque(["Resources", "MyAppStreamFleet", "CreationPolicy"]),
            },
            [
                ValidationError(
                    "{} is not of type 'boolean'",
                    rule=CreationPolicy(),
                    path=deque(["StartFleet", "Type"]),
                    schema_path=deque(
                        ["properties", "StartFleet", "properties", "Type", "type"]
                    ),
                    validator="type",
                )
            ],
        ),
        (
            "Valid ASG",
            {
                "AutoScalingCreationPolicy": {"MinSuccessfulInstancesPercent": 100},
                "ResourceSignal": {"Count": 1, "Timeout": "60"},
            },
            {
                "path": deque(["Resources", "MyAutoScalingGroup", "CreationPolicy"]),
            },
            [],
        ),
        (
            "Invalid ASG",
            {
                "AutoScalingCreationPolicy": {"MinSuccessfulInstancesPercent": 100},
                "ResourceSignal": {"Count": "one", "Timeout": "60"},
            },
            {
                "path": deque(["Resources", "MyAutoScalingGroup", "CreationPolicy"]),
            },
            [
                ValidationError(
                    "'one' is not of type 'integer'",
                    rule=CreationPolicy(),
                    path=deque(["ResourceSignal", "Count"]),
                    schema_path=deque(
                        ["properties", "ResourceSignal", "properties", "Count", "type"]
                    ),
                    validator="type",
                )
            ],
        ),
        (
            "Valid Wait Condition",
            {"ResourceSignal": {"Timeout": "PT15M", "Count": "5"}},
            {
                "path": deque(["Resources", "MyWaitCondition", "CreationPolicy"]),
            },
            [],
        ),
        (
            "Invalid Wait Condition",
            {"ResourceSignal": {"Timeout": "PT15M", "Count": "five"}},
            {
                "path": deque(["Resources", "MyWaitCondition", "CreationPolicy"]),
            },
            [
                ValidationError(
                    "'five' is not of type 'integer'",
                    rule=CreationPolicy(),
                    path=deque(["ResourceSignal", "Count"]),
                    schema_path=deque(
                        ["properties", "ResourceSignal", "properties", "Count", "type"]
                    ),
                    validator="type",
                )
            ],
        ),
        (
            "Valid instance",
            {"ResourceSignal": {"Timeout": "PT15M", "Count": "5"}},
            {
                "path": deque(["Resources", "MyWaitCondition", "CreationPolicy"]),
            },
            [],
        ),
        (
            "Invalid Instance",
            {"Foo": {"Bar"}},
            {
                "path": deque(["Resources", "MyInstance", "CreationPolicy"]),
            },
            [],
        ),
        (
            "Wait condition ignored on wrong type",
            {"Foo": {"Bar"}},
            {
                "path": deque(["Resources", "MyLambdaFunction", "CreationPolicy"]),
            },
            [],
        ),
        (
            "Invalid but integer name ",
            {"ResourceSignal": {"Timeout": "PT15M", "Count": "five"}},
            {
                "path": deque(["Resources", 1, "CreationPolicy"]),
            },
            [],
        ),
    ],
    indirect=["path"],
)
def test_deletion_policy(name, instance, expected, rule, validator):

    rule = CreationPolicy()
    errors = list(
        rule.validate(
            validator=validator,
            dP="creationpolicy",
            instance=instance,
            schema={},
        )
    )

    assert errors == expected, f"{name}: {errors} != {expected}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class Type(CloudFormationLintRule):
    """Check if Outputs have the correct type"""

    id = "E6003"
    shortdesc = "Check the type of Outputs"
    description = "Validate the type of properties in the Outputs section"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html"
    tags = ["outputs"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from pathlib import Path
from test.integration import BaseCliTestCase

from cfnlint import ConfigMixIn


class TestQuickStartTemplates(BaseCliTestCase):
    """Test QuickStart Templates Parsing"""

    # ruff: noqa: E501
    scenarios = [
        {
            "filename": "test/fixtures/templates/good/generic.yaml",
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": "test/fixtures/templates/good/minimal.yaml",
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": "test/fixtures/templates/good/transform.yaml",
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": "test/fixtures/templates/issues/sam_w_conditions.yaml",
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": (
                "test/fixtures/templates/bad/transform_serverless_template.yaml"
            ),
            "results": [
                {
                    "Filename": "test/fixtures/templates/bad/transform_serverless_template.yaml",
                    "Id": "9e05773a-b0d0-f157-2955-596d9bd54749",
                    "Level": "Error",
                    "Location": {
                        "End": {"ColumnNumber": 2, "LineNumber": 1},
                        "Path": None,
                        "Start": {"ColumnNumber": 1, "LineNumber": 1},
                    },
                    "Message": "Error transforming template: Resource with id [myFunctionMyTimer] is invalid. Missing required property 'Schedule'.",
                    "ParentId": None,
                    "Rule": {
                        "Description": "Errors found when performing transformation on the template",
                        "Id": "E0001",
                        "ShortDescription": "Error found when transforming the template",
                        "Source": "https://github.com/aws-cloudformation/cfn-lint",
                    },
                },
                {
                    "Filename": "test/fixtures/templates/bad/transform_serverless_template.yaml",
                    "Id": "fd751fa3-7d1f-e194-7108-eb08352814c8",
                    "Level": "Error",
                    "Location": {
                        "End": {"ColumnNumber": 2, "LineNumber": 1},
                        "Path": None,
                        "Start": {"ColumnNumber": 1, "LineNumber": 1},
                    },
                    "Message": "Error transforming template: Resource with id [ExampleLayer] is invalid. Missing required property 'ContentUri'.",
                    "ParentId": None,
                    "Rule": {
                        "Description": "Errors found when performing transformation on the template",
                        "Id": "E0001",
                        "ShortDescription": "Error found when transforming the template",
                        "Source": "https://github.com/aws-cloudformation/cfn-lint",
                    },
                },
                {
                    "Filename": "test/fixtures/templates/bad/transform_serverless_template.yaml",
                    "Id": "74181426-e865-10eb-96fd-908dfd30a358",
                    "Level": "Error",
                    "Location": {
                        "End": {"ColumnNumber": 2, "LineNumber": 1},
                        "Path": None,
                        "Start": {"ColumnNumber": 1, "LineNumber": 1},
                    },
                    "Message": "Error transforming template: Resource with id [AppName] is invalid. Resource is missing the required [Location] property.",
                    "ParentId": None,
                    "Rule": {
                        "Description": "Errors found when performing transformation on the template",
                        "Id": "E0001",
                        "ShortDescription": "Error found when transforming the template",
                        "Source": "https://github.com/aws-cloudformation/cfn-lint",
                    },
                },
            ],
            "exit_code": 2,
        },
        {
            "filename": "test/fixtures/templates/good/conditions.yaml",
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": "test/fixtures/templates/good/resources_codepipeline.yaml",
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": "test/fixtures/templates/good/resources_cognito_userpool_tag_is_string_map.yaml",
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": "test/fixtures/templates/bad/resources_cognito_userpool_tag_is_list.yaml",
            "results": [
                {
                    "Filename": str(
                        Path(
                            "test/fixtures/templates/bad/resources_cognito_userpool_tag_is_list.yaml"
                        )
                    ),
                    "Id": "3732b0a0-6d44-72af-860a-88e5f8ca790c",
                    "Level": "Error",
                    "Location": {
                        "Start": {"ColumnNumber": 7, "LineNumber": 16},
                        "End": {"ColumnNumber": 19, "LineNumber": 16},
                        "Path": [
                            "Resources",
                            "MyCognitoUserPool",
                            "Properties",
                            "UserPoolTags",
                        ],
                    },
                    "Message": (
                        "[{'Key': 'Key1', 'Value': 'Value1'}, {'Key': 'Key2', 'Value':"
                        " 'Value2'}] is not of type 'object'"
                    ),
                    "ParentId": None,
                    "Rule": {
                        "Description": (
                            "Checks resource property values with Primitive Types for"
                            " values that match those types."
                        ),
                        "Id": "E3012",
                        "ShortDescription": "Check resource properties values",
                        "Source": "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#type",
                    },
                }
            ],
            "exit_code": 2,
        },
        {
            "filename": "test/fixtures/templates/good/transform_serverless_api.yaml",
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": (
                "test/fixtures/templates/good/transform_serverless_function.yaml"
            ),
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": (
                "test/fixtures/templates/good/transform_serverless_globals.yaml"
            ),
            "results": [
                {
                    "Filename": str(
                        Path(
                            "test/fixtures/templates/good/transform_serverless_globals.yaml"
                        )
                    ),
                    "Id": "f0f6c586-81bc-9182-de02-659a3a1a5b2c",
                    "Level": "Error",
                    "Location": {
                        "End": {"ColumnNumber": 13, "LineNumber": 10},
                        "Path": ["Resources", "myFunction", "Properties", "Runtime"],
                        "Start": {"ColumnNumber": 3, "LineNumber": 10},
                    },
                    "Message": "Runtime 'nodejs6.10' was deprecated on '2019-08-12'. Creation was disabled on '2019-08-12' and update on '2019-08-12'. Please consider updating to 'nodejs20.x'",
                    "ParentId": None,
                    "Rule": {
                        "Description": (
                            "Check if an EOL Lambda Runtime is specified and you cannot update the function"
                        ),
                        "Id": "E2533",
                        "ShortDescription": (
                            "Check if Lambda Function Runtimes are updatable"
                        ),
                        "Source": "https://docs.aws.amazon.com/lambda/latest/dg/runtime-support-policy.html",
                    },
                }
            ],
            "exit_code": 2,
        },
        {
            "filename": "test/fixtures/templates/good/transform/list_transform.yaml",
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": (
                "test/fixtures/templates/good/transform/list_transform_many.yaml"
            ),
            "results": [],
            "exit_code": 0,
        },
        {
            "filename": (
                "test/fixtures/templates/good/transform/list_transform_not_sam.yaml"
            ),
            "results": [],
            "exit_code": 0,
        },
    ]

    def test_templates(self):
        """Test Successful JSON Parsing"""
        self.run_scenarios()

    def test_module_integration(self):
        """Test same templates using integration approach"""
        self.run_module_integration_scenarios(
            ConfigMixIn([], include_checks=["W", "E"])
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ecs.FargateCpuMemory import FargateCpuMemory


@pytest.fixture(scope="module")
def rule():
    rule = FargateCpuMemory()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 256,
                "Memory": "512",
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": "512",
                "Memory": 1024,
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": ".5 vCPU",
                "Memory": "1 GB",
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": "1024",
                "Memory": "2048",
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 2048,
                "Memory": 4096,
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": "4 vcpu",
                "Memory": 30720,
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": "4 vcpu",
                "Memory": "22 gb",
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 8192,
                "Memory": "16 GB",
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": "16     vCpU",
                "Memory": 122880,
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 16384,
                "Memory": 123904,
            },
            [
                ValidationError(
                    "Cpu 16384 is not compatible with memory 123904",
                    rule=FargateCpuMemory(),
                    path=deque([]),
                    validator="anyOf",
                    schema_path=deque(["then", "anyOf"]),
                )
            ],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 16384,
                "Memory": "1 GB",
            },
            [
                ValidationError(
                    "Cpu 16384 is not compatible with memory '1 GB'",
                    rule=FargateCpuMemory(),
                    path=deque([]),
                    validator="anyOf",
                    schema_path=deque(["then", "anyOf"]),
                )
            ],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE", "Foo"],
                "Cpu": 4096,
                "Memory": 512,
            },
            [
                ValidationError(
                    "Cpu 4096 is not compatible with memory 512",
                    rule=FargateCpuMemory(),
                    path=deque([]),
                    validator="anyOf",
                    schema_path=deque(["then", "anyOf"]),
                )
            ],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE", "Foo"],
                "Cpu": 4096,
                "Memory": 16385,
            },
            [
                ValidationError(
                    "Cpu 4096 is not compatible with memory 16385",
                    rule=FargateCpuMemory(),
                    path=deque([]),
                    validator="anyOf",
                    schema_path=deque(["then", "anyOf"]),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.functions._BaseFn import BaseFn


class _ChildRule(CloudFormationLintRule):
    id = "XXXXX"


@pytest.fixture(scope="module")
def rule():
    rule = BaseFn(resolved_rule="XXXXX")
    rule.child_rules["XXXXX"] = _ChildRule()
    yield rule


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Dynamic references are ignored",
            {"Fn::Sub": "{{resolve:ssm:${AWS::AccountId}/${AWS::Region}/ac}}"},
            {"enum": ["Foo"]},
            [],
        ),
        ("Everything is fine", {"Fn::Sub": "Foo"}, {"enum": ["Foo"]}, []),
        (
            "Resolved Fn::Sub has no strict type validation",
            {"Fn::Sub": "2"},
            {"type": ["integer"]},
            [],
        ),
        (
            "Standard error",
            {"Fn::Sub": "Bar"},
            {"enum": ["Foo"]},
            [
                ValidationError(
                    message=(
                        "{'Fn::Sub': 'Bar'} is not one of "
                        "['Foo'] when '' is resolved"
                    ),
                    path=deque(["Fn::Sub"]),
                    validator="",
                    schema_path=deque(["enum"]),
                    rule=_ChildRule(),
                )
            ],
        ),
    ],
)
def test_resolve(name, instance, schema, expected, validator, rule):
    errs = list(rule.resolve(validator, schema, instance, {}))
    assert errs == expected, f"{name!r} failed and got errors {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.context.context import Path
from cfnlint.jsonschema import CfnTemplateValidator
from cfnlint.jsonschema._filter import FunctionFilter
from cfnlint.template import Template


@pytest.fixture(scope="module")
def filter():
    filter = FunctionFilter()
    yield filter


@pytest.fixture()
def template():
    template = Template(
        None,
        {
            "Parameters": {"MyParameter": {"Type": "String"}},
            "Conditions": {"Condition": {"Fn::Equals": [{"Ref": "MyParameter"}, ""]}},
        },
        ["us-east-1"],
    )
    yield template


@pytest.mark.parametrize(
    "name,instance,schema,path,functions,expected",
    [
        (
            "Don't validate dynamic references inside of function",
            "{{resolve:ssm:${AWS::AccountId}/${AWS::Region}/ac}}",
            {"enum": "Foo"},
            deque(["Foo", "Test", "Fn::Sub"]),
            [],
            [],
        ),
        (
            "Validate dynamic references",
            "{{resolve:ssm:secret}}",
            {"enum": "Foo"},
            deque(["Foo", "Test"]),
            [],
            [
                ("{{resolve:ssm:secret}}", {"dynamicReference": {"enum": "Foo"}}),
            ],
        ),
        (
            "Lack of functions returns the schema and instance",
            {
                "Foo": {"Ref": "AWS::NoValue"},
            },
            {"required": ["Foo"]},
            deque([]),
            [],
            [
                (
                    {
                        "Foo": {"Ref": "AWS::NoValue"},
                    },
                    {"required": ["Foo"]},
                ),
            ],
        ),
        (
            "Filtered schemas",
            {
                "Foo": {"Ref": "AWS::NoValue"},
            },
            {"required": ["Foo"]},
            deque([]),
            ["Ref", "Fn::If"],
            [
                (
                    {},
                    {"required": ["Foo"]},
                ),
                ({"Foo": {"Ref": "AWS::NoValue"}}, {"cfnLint": [""]}),
            ],
        ),
        (
            "Filtered schemas with list",
            [
                {
                    "Fn::If": [
                        "Condition",
                        {"Key": "1", "Value": "A"},
                        {"Ref": "AWS::NoValue"},
                    ]
                },
                {
                    "Key": "2",
                    "Value": "B",
                },
                {"Ref": "AWS::NoValue"},
            ],
            {"items": {"required": ["Key"]}},
            deque([]),
            ["Ref", "Fn::If"],
            [
                (
                    [
                        {
                            "Fn::If": [
                                "Condition",
                                {"Key": "1", "Value": "A"},
                                {"Ref": "AWS::NoValue"},
                            ]
                        },
                        {
                            "Key": "2",
                            "Value": "B",
                        },
                        {"Ref": "AWS::NoValue"},
                    ],
                    {"items": {"required": ["Key"]}, "cfnLint": [""]},
                ),
            ],
        ),
    ],
)
def test_filter(
    name,
    instance,
    schema,
    path,
    functions,
    expected,
    filter,
    template,
):
    context = create_context_for_template(template)
    validator = CfnTemplateValidator(
        context=context,
        cfn=template,
        schema=schema,
    )
    validator = validator.evolve(
        context=validator.context.evolve(
            path=Path(path),
            functions=functions,
        )
    )

    results = list(filter.filter(validator, instance, schema))

    assert len(results) == len(expected), f"For test {name} got {len(results)} results"

    for result, (exp_instance, exp_schema) in zip(results, expected):
        assert result[0] == exp_instance, f"For test {name} got {result[0]!r}"
        assert result[1] == exp_schema, f"For test {name} got {result[1]!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.jsonschema._keywords import enum
from cfnlint.rules import CloudFormationLintRule


class Enum(CloudFormationLintRule):
    """Check if properties have a valid value"""

    id = "E3030"
    shortdesc = "Check if properties have a valid value"
    description = "Check if properties have a valid value in case of an enumator"
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#enum"
    tags = ["resources", "property", "allowed value"]
    child_rules = {
        "W2030": None,
    }

    def enum(self, validator, enums, instance, schema):
        if (
            len(validator.context.path.value_path) > 0
            and validator.context.path.value_path[0] == "Parameters"
        ):
            if self.child_rules.get("W2030"):
                yield from self.child_rules["W2030"].enum(
                    validator, enums, instance, schema
                )
            return
        yield from enum(validator, enums, instance, schema)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import regex as re

from cfnlint.helpers import ensure_list, is_function
from cfnlint.jsonschema import ValidationError, ValidationResult
from cfnlint.jsonschema.protocols import Validator
from cfnlint.rules.jsonschema import CfnLintKeyword


class PermissionSourceAccount(CfnLintKeyword):

    id = "W3663"
    shortdesc = "Validate SourceAccount is required property"
    description = (
        "When configuration a Lambda permission with a SourceArn "
        "that doesn't have an AccountId you should also specify "
        "the SourceAccount"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-lambda-permission.html#cfn-lambda-permission-sourceaccount"
    tags = ["resources", "lambda", "permission"]

    def __init__(self):
        super().__init__(
            keywords=["Resources/AWS::Lambda::Permission/Properties"],
        )

    def _validate_is_gettatt_to_bucket(self, validator: Validator, value: Any) -> bool:
        value = ensure_list(value)[0].split(".")[0]

        resource = validator.context.resources[value]
        if resource.type == "AWS::S3::Bucket":
            return True
        return False

    def validate(
        self, validator: Validator, _: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not isinstance(instance, dict):
            return

        for scenario in validator.cfn.get_object_without_conditions(
            instance, ["SourceArn", "SourceAccount"]
        ):
            if scenario.get("Scenario"):
                scenario_validator = validator.evolve(
                    context=validator.context.evolve(
                        conditions=validator.context.conditions.evolve(
                            status=scenario.get("Scenario")
                        )
                    )
                )
            else:
                scenario_validator = validator.evolve()

            source_arn = scenario.get("Object").get("SourceArn")
            source_account = scenario.get("Object").get("SourceAccount")
            if not source_arn:
                continue

            if isinstance(source_arn, str):
                if re.search(r":\d{12}:", source_arn):
                    continue

            fn_k, fn_v = is_function(source_arn)
            if fn_k is not None:
                if fn_k == "Fn::GetAtt":
                    if not self._validate_is_gettatt_to_bucket(
                        scenario_validator, fn_v
                    ):
                        continue
                else:
                    continue

            if not source_account:
                yield ValidationError(
                    "'SourceAccount' is a required property",
                    validator="required",
                    rule=self,
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_ecs_taskdefinition
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class TaskDefinitionEssentialContainer(CfnLintJsonSchema):
    """
    Check ECS TaskDefinition ContainerDefinitions Property
    Specifies at least one Essential Container
    """

    id = "E3042"
    shortdesc = "Validate at least one essential container is specified"
    description = (
        "Check that every TaskDefinition specifies at least one essential container"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ecs-taskdefinition-containerdefinitions.html#cfn-ecs-taskdefinition-containerdefinition-essential"
    tags = ["properties", "ecs", "task", "container", "fargate"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::ECS::TaskDefinition/Properties/ContainerDefinitions/*"
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_ecs_taskdefinition,
                filename="containerdefinitions_essential.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return "At least one essential container is required"
